# import pandas as pd
import os
import pickle
from pprint import pprint
from datetime import datetime
import numpy as np
from collections import defaultdict

from .utils.infrastructure import writeGrainList2File, readPickleFile2GrainUnique, specialTokens, UNK_ID
from .utils.infrastructure import modify_wordBoundary_with_hyperBoundary, trans_charLabels_to_wordLabels
from .utils.infrastructure import getTagDict, trans_bioesTag

from .utils.pyramid import CorpusGroupsReader, FolderTextsReaders, segText2Sents, segSent2Tokens
from .utils.anno import getCITText, getCITSents, getSSET_from_CIT
from .utils.vocab import buildTokens, get_GU_or_LKP, get_num_freq

from .utils.channel import CONTEXT_IND_CHANNELS, CONTEXT_DEP_CHANNELS, ANNO_CHANNELS, CHANNEL_ABBR
from .utils.channel import Channel_Dep_Methods, Channel_Dep_TagSets, getChannelName, get_Channel_Settings


# labels and tags
# labels: ['Person', 'Location']
# tags  : ['O', Person-B', 'Person-I', 'Location-B', 'Location-I']

# token, token_idx, token_locidx
# "New",        10,          123
# DTU['New'] = 10
# token(123) = "New", perhaps there is another token_locidx which is also "New".

def utf8len(s):
    return len(s.encode('utf-8'))


class BasicObject(object):
    # basic path
    Base_Dir = None

    # 五大老五奉行
    CORPUS = {}
    GROUP  = {}
    TEXT   = {}
    SENT   = {}
    TOKEN  = {}

    # token and hyper fields
    TokenVocab = None
    VOCAB = {}
    # deal with transfer hyper fields tokens
    TRANS = {}
    # deal with the sub field information
    LOOKUP = {}
    FREQ   = {}

    CTX_DEP_TMP = {}

    # other items
    specialTokens = specialTokens
    CONTEXT_IND_CHANNELS = CONTEXT_IND_CHANNELS
    CONTEXT_DEP_CHANNELS = CONTEXT_DEP_CHANNELS
    ANNO_CHANNELS        = ANNO_CHANNELS
    CHANNEL_ABBR         = CHANNEL_ABBR

    @classmethod
    def INIT(cls, CORPUSPath, 
             Corpus2GroupMethod, 
             Group2TextMethod, 
             Text2SentMethod, 
             Sent2TokenMethod, TOKENLevel, min_token_freq = 1,
             use_hyper = [], Channel_Dep_Methods = Channel_Dep_Methods, Channel_Dep_TagSets = Channel_Dep_TagSets, 
             anno = False, anno_keywords = {}):
        
        Channel_Dep_Methods = {i:k for i,k in Channel_Dep_Methods.items() if i in use_hyper} 
        Channel_Dep_TagSets = {i:k for i,k in Channel_Dep_TagSets.items() if i in use_hyper} 
        
        ################################################################################################################
        corpus_name = CORPUSPath.split('/')[-2]
        Data_Dir = os.path.join('data', corpus_name, TOKENLevel)
        File_Dir = os.path.join(Data_Dir, 'Pyramid', '_file')
        cls.Data_Dir = Data_Dir
        if not os.path.exists(File_Dir):
            os.makedirs(File_Dir)


        cls.Channel_Hyper_Path = {}
        cls.Channel_Hyper_Path['token'] = os.path.join(File_Dir, 'token.txt')
        if os.path.isfile(cls.Channel_Hyper_Path['token']):
            os.remove(cls.Channel_Hyper_Path['token'])

        ################################################################################################################
        # for hyper field
        Path_Key = os.path.join(Data_Dir, 'Vocab')
        cls.VOCAB = {}
        cls.VOCAB[Path_Key] = {}
        for ch, tagSet in Channel_Dep_TagSets.items():
            LGU = getTagDict(tagSet, tagScheme = 'BIOES')
            DGU = dict(zip(LGU, range(len(LGU))))
            channel_name = ch + '-bioes'
            cls.VOCAB[Path_Key][channel_name] = (LGU, DGU)
            cls.Channel_Hyper_Path[ch] =  os.path.join(File_Dir, channel_name + '.txt')
            if os.path.isfile(cls.Channel_Hyper_Path[ch]):
                os.remove(cls.Channel_Hyper_Path[ch])

        # to enrich annoE vocab inside it.
        if anno: 
            cls.VOCAB[Path_Key]['annoE-bioes'] = (['O'], {'O': 0}); anno_labels = []
            cls.Channel_Hyper_Path['annoE'] =  os.path.join(File_Dir, 'annoE-bioes' + '.txt')
            if os.path.isfile(cls.Channel_Hyper_Path['annoE']):
                os.remove(cls.Channel_Hyper_Path['annoE'])

        ################################################################################################################
        # for annotation
        # assert anno == False or '.' in anno or anno == 'embed'
    
        ################################################################################################################
        # init 五大老五奉行
        # corpus information
        CORPUS = {}; CORPUS['CORPUSPath'] = CORPUSPath; CORPUS['Data_Dir'] = Data_Dir; CORPUS['EndIDXGroups'] = []
        
        # group information
        GROUP = {}; GROUP['Corpus2GroupMethod'] = Corpus2GroupMethod 
        GROUP['GroupType']   = 'File' if Corpus2GroupMethod != 'Dir' else 'Dir'; GROUP['group_names'] = []; GROUP['EndIDXTexts'] = []
        
        # text information
        TEXT = {}; TEXT['Group2TextMethod'] = Group2TextMethod; TEXT['EndIDXSents'] = []
        if Group2TextMethod == 'file': TEXT['ORIGFileName'] = []
        if 'ANNOIden' in anno_keywords: TEXT['ANNOFileName'] = []
            
        # sentence information
        # TODO: change these to array to save space
        SENT = {}; SENT['Text2SentMethod'] = Text2SentMethod; SENT['EndIDXTokens'] = np.zeros(5000000, dtype = int)
        # even more file information will be included here. TODO change this to numpy
        for ch, path in cls.Channel_Hyper_Path.items(): SENT[path] = np.zeros(5000000, dtype = int)
        
        # token information
        TOKEN = {}; TOKEN['Sent2TokenMethod'] = Sent2TokenMethod
        TOKEN['TOKENLevel'] = TOKENLevel; TOKEN['Channel_Hyper_Path'] = cls.Channel_Hyper_Path


        # consider how to deal with the annotation information
        ANNO = {}; ANNO['anno'] = anno; ANNO['anno_keywords'] = anno_keywords

        ################################################################################################################
        CorpusGroups, GroupType = CorpusGroupsReader(CORPUSPath, iden = Corpus2GroupMethod)
        # print(CorpusGroups, GroupType)
        assert GROUP['GroupType'] == GroupType
        pprint(GroupType)

        oldDTU = defaultdict(int)# change this to default dict with int
        oldLTU = []
        oldidx2freq = []
        token_num_in_corpus = 0

        total_sent_num = 0
        for group_idx, group_name in enumerate(CorpusGroups):
            # the following block deals with each group in a corpus
            print(group_name)
            text_names = CorpusGroups[group_name]
            GroupTexts = FolderTextsReaders[Group2TextMethod](group_name, text_names, anno, **anno_keywords)
            textIdx = 0
            for strText_SSET_O_A in GroupTexts:
                # the following block deals with each text in a group
                strText, SSETText, origTextName, annoTextName = strText_SSET_O_A
                
                # print(strText, SSETText, origTextName, annoTextName)
                
                strSents = segText2Sents(strText, method = Text2SentMethod) 

                # skip the sents that are empty and the sents which doesn't have annotations.
                # if len(strSents) == 0 or (anno and len(SSETText) == 0): continue
                if len(strSents) == 0: continue

                if anno: tokenizedSents = []
                # the following to blocks deal with multiple hyperfields (include token and annoE) for each strSent in strSents
                # the main input are strText and strSents, (especially strSents)
                
                # block 1: for token and other hyper fields.
                for idx, strSent in enumerate(strSents):
                    # the following block deals with each strSent in a text.
                    # if strTags is not None, strTokens and strTags share the same length, 
                    # and this should be assert inside segSent2Tokens
                    
                    # should assert [' ' not in strToken for strToken in strTokens]
                    # and len(hyper_info) == len(strTokens)

                    strTokens, hyper_info = segSent2Tokens(strSent, Sent2TokenMethod, TOKENLevel, Channel_Dep_Methods)

                    # LESSION: update strSents, which will be used in annotation.
                    if anno: tokenizedSents.append(strTokens)
                    # deal with tokens
                    for token in strTokens:
                        # the following block deals with each token in a text.
                        if token in oldDTU:
                            # deal with old words
                            token_idx = oldDTU[token]
                            oldidx2freq[token_idx] += 1
                        else:
                            # deal with new words
                            token_idx  = len(oldDTU)
                            oldDTU[token] = token_idx
                            oldidx2freq.append(1)
                            oldLTU.append(token)
                        token_num_in_corpus = token_num_in_corpus + 1

                    with open(cls.Channel_Hyper_Path['token'], 'a') as f:
                        line_sentence = ' '.join(strTokens) + '\n'
                        f.write(line_sentence)
                        bytelenSent = utf8len(line_sentence)
                    
                    # total_sent_num increase all the time
                    # if it is 0, then SENT[cls.Channel_Hyper_Path['token']][total_sent_num - 1] will return 0
                    SENT[cls.Channel_Hyper_Path['token']][total_sent_num] = SENT[cls.Channel_Hyper_Path['token']][total_sent_num - 1] + bytelenSent
                    # try:
                    #     # SENT[cls.Channel_Hyper_Path['token']].append(SENT[cls.Channel_Hyper_Path['token']][-1] + bytelenSent)
                    # except:
                    #     SENT[cls.Channel_Hyper_Path['token']][total_sent_num] = bytelenSent

                    # if use word level, make the sure that the return grains are the same.
                    # this is for generating the hyper field information.
                    # before using the hyper field information, we must know their labels before.
                    for ch, ch_grain_sent in hyper_info.items():
                        ch_grain_sent = [str(cls.VOCAB[Path_Key][channel_name][1].get(i)) for i in ch_grain_sent]
                        assert len(ch_grain_sent) == len(strTokens)
                        # or you can create a file, which is LineSentence type file. each line is a sentence.
                        with open(cls.Channel_Hyper_Path[ch], 'a') as f:
                            line_sentence = ' '.join(ch_grain_sent) + '\n'
                            f.write(line_sentence)
                            bytelenSent = utf8len(line_sentence)

                        SENT[cls.Channel_Hyper_Path[ch]][total_sent_num] = SENT[cls.Channel_Hyper_Path[ch]][total_sent_num - 1] + bytelenSent
                        # try:
                        #     SENT[cls.Channel_Hyper_Path[ch]].append(SENT[cls.Channel_Hyper_Path[ch]][-1] + bytelenSent)
                        # except:
                        #     SENT[cls.Channel_Hyper_Path[ch]].append(bytelenSent)

                    # adding the sentence length, i.e., the token number in this sentence.
                    lenSent = len(strTokens)
                    SENT['EndIDXTokens'][total_sent_num] = SENT['EndIDXTokens'][total_sent_num - 1] + lenSent
                    # try:
                    #     SENT['EndIDXTokens'].append(SENT['EndIDXTokens'][-1] + lenSent)
                    # except:
                    #     SENT['EndIDXTokens'].append(lenSent)

                    # add a num to total_sent_num
                    # to process next sentence
                    total_sent_num = total_sent_num + 1


                    # logging the information and epxand the length if necessary
                    if total_sent_num % 100000 == 0:
                        print(total_sent_num, '--', lenSent, datetime.now(), '\t',SENT['EndIDXTokens'][total_sent_num-1],  '\t', len(oldidx2freq), len(oldDTU), len(oldLTU))
                        print('\t', len(SENT['EndIDXTokens']), 'token')
                        for ch, ch_grain_sent in hyper_info.items():
                            print('\t', len(SENT[cls.Channel_Hyper_Path[ch]]), ch)# .append(bytelenSent)

                        if len(SENT['EndIDXTokens']) - total_sent_num <= 100000:
                            print('\tEnlarging the SENT size...')
                            print('\tcurrent leng is:', len(SENT['EndIDXTokens']))
                            SENT[cls.Channel_Hyper_Path['token']] = np.concatenate([SENT[cls.Channel_Hyper_Path['token']], np.zeros(2000000, dtype = int)])
                            SENT['EndIDXTokens'] = np.concatenate([SENT['EndIDXTokens'], np.zeros(2000000, dtype = int)])
                            for ch, ch_grain_sent in hyper_info.items():
                                SENT[cls.Channel_Hyper_Path[ch]] = np.concatenate([SENT[cls.Channel_Hyper_Path[ch]], np.zeros(2000000, dtype = int)])
                            print('\tupdated leng is:', len(SENT['EndIDXTokens']))


                # block1: for annotation.
                if anno: 
                    # if SSETText == []: print('\nThe SSET of this Text is Empty!!!', '\n', strText, '\n')

                    # produce the VOCAB
                    new_labels = set([sset[-1] for sset in SSETText])
                    for label in new_labels:
                        if label not in anno_labels:
                            anno_labels.append(label)
                            for suffix in ['-B', '-I', '-E', '-S']:
                                cls.VOCAB[Path_Key]['annoE-bioes'][1][label + suffix] = len(cls.VOCAB[Path_Key]['annoE-bioes'][1])
                                cls.VOCAB[Path_Key]['annoE-bioes'][0].append(label + suffix)

                    # it will check strText and SSET inside getCITText
                    CITText  = getCITText(strText, SSETText, TOKENLevel) 
                    
                    # get CITSents
                    CITSents = getCITSents(tokenizedSents, CITText, TOKENLevel)
                    
                    # pay attention: do not change total_sent_num
                    last_total_sent_num = total_sent_num - len(CITSents) 
                    for sentIdx, CITSent in enumerate(CITSents):
                        anno_tags = [CITToken[2] for CITToken in CITSent]
                        anno_tags = [str(cls.VOCAB[Path_Key]['annoE-bioes'][1][i]) for i in anno_tags]

                        # make sure anno seq is equal to token seq
                        assert len(anno_tags) == len(tokenizedSents[sentIdx])
                        
                        with open(cls.Channel_Hyper_Path['annoE'], 'a') as f:
                            line_sentence = ' '.join(anno_tags) + '\n'
                            f.write(line_sentence)
                            bytelenSent = utf8len(line_sentence)
                        
                        sent_loc_id = last_total_sent_num + sentIdx
                        SENT[cls.Channel_Hyper_Path['annoE']][sent_loc_id] = SENT[cls.Channel_Hyper_Path['annoE']][sent_loc_id - 1] + bytelenSent
                        # try:
                        #     SENT[cls.Channel_Hyper_Path['annoE']].append(SENT[cls.Channel_Hyper_Path['annoE']][-1] + bytelenSent)
                        # except:
                        #     SENT[cls.Channel_Hyper_Path['annoE']].append(bytelenSent)
                    assert sent_loc_id + 1 == total_sent_num
                    if annoTextName:
                        TEXT['ANNOFileName'].append(annoTextName)
                    
                # deal with the text and its sentences information
                lenText = len(strSents)
                try:
                    TEXT['EndIDXSents'].append(TEXT['EndIDXSents'][-1] + lenText)
                except:
                    TEXT['EndIDXSents'].append(lenText)
                    
                if origTextName:
                    TEXT['ORIGFileName'].append(origTextName)
                
                textIdx = textIdx + 1
                # if textIdx % 100000 == 0:
                #     print(textIdx, '--', len(strSents), len(strText), datetime.now())
                #     print(len(SENT['EndIDXTokens']))
                #     for ch, ch_grain_sent in hyper_info.items():
                #         print(len(SENT[cls.Channel_Hyper_Path[ch]]), ch)# .append(bytelenSent)

            # Back to Folder
            lenGroup = textIdx
            GROUP['group_names'].append(group_name)
            try:
                GROUP['EndIDXTexts'].append(GROUP['EndIDXTexts'][-1] + lenGroup)
            except:
                GROUP['EndIDXTexts'].append(lenGroup)
        
        # if there is an error, folderIdx will be referenced before assignment
        lenCorpus = group_idx + 1 
        CORPUS['EndIDXGroups'] = [lenCorpus]

        ####################################################################################### 
        # print('reorder the frequency ...')
        del oldDTU

        idx2freq = np.sort(oldidx2freq)[::-1]
        newidx2oldidx = np.argsort(oldidx2freq)[::-1]
        del oldidx2freq
        
        # oldidx2newidx = np.zeros(len(newidx2oldidx), dtype= int)
        # for new_idx, old_idx in enumerate(newidx2oldidx):
        #     oldidx2newidx[old_idx] = new_idx
        # del oldidx2newidx

        LTU = []
        for new_idx in range(len(oldLTU)):
            LTU.append(oldLTU[newidx2oldidx[new_idx]])
        del oldLTU
        del newidx2oldidx

        DTU = defaultdict(int)
        for new_idx, token in enumerate(LTU):
            DTU[token] = new_idx
        
        print('Total Num of All    Tokens', token_num_in_corpus)
        print('Total Num of Unique Tokens', len(LTU))

        # load 五大老五奉行, TODO: this can be changed
        CORPUS['EndIDXGroups'] = np.array(CORPUS['EndIDXGroups'])
        CORPUS['length']       = len(CORPUS['EndIDXGroups'])

        GROUP['EndIDXTexts']   = np.array(GROUP['EndIDXTexts'])
        GROUP['length']        = len(GROUP['EndIDXTexts'])

        TEXT['EndIDXSents']    = np.array(TEXT['EndIDXSents'])
        TEXT['length']         = len(TEXT['EndIDXSents'])

        SENT['EndIDXTokens']   = SENT['EndIDXTokens'][:total_sent_num]
        SENT['length']         = len(SENT['EndIDXTokens'])
        for ch, path in cls.Channel_Hyper_Path.items():
            SENT[path] = SENT[path][:total_sent_num]

        TOKEN['length']        = token_num_in_corpus
        
        cls.CORPUS = CORPUS
        cls.GROUP  = GROUP
        cls.TEXT   = TEXT
        cls.SENT   = SENT
        cls.TOKEN  = TOKEN
        
        cls.idx2freq  = idx2freq
        cls.TokenVocab = (LTU, DTU)
        orig_vocab_length = len(cls.TokenVocab[0])
        cls.original_vocab_token_num = orig_vocab_length
        cls.current_vocab_token_num  = orig_vocab_length
        cls.OBJECT_TO_PICKLE()

        max_vocab_token_num = get_num_freq(cls.idx2freq, min_token_freq = min_token_freq)
        cls.min_token_freq = min_token_freq
        # cls.max_vocab_token_num = max_vocab_token_num
        cls._buildGVforToken(max_vocab_token_num, min_token_freq)

        folder = 'F' + str(min_token_freq) 
        Path_Key = os.path.join(cls.Data_Dir, 'Vocab', folder)
        if not os.path.exists(Path_Key):
            os.makedirs(Path_Key)


    @classmethod
    def OBJECT_TO_PICKLE(cls):
        Data_Dir = cls.Data_Dir
        
        ################################################################################
        Pyramid_Dir = os.path.join(Data_Dir, 'Pyramid')
        if not os.path.exists(Pyramid_Dir):
            os.makedirs(Pyramid_Dir)
        d = {'CORPUS': cls.CORPUS, 
             'GROUP':  cls.GROUP, 
             'TEXT':   cls.TEXT, 
             'SENT':   cls.SENT, 
             'TOKEN':  cls.TOKEN}
        for k, v in d.items():
            pickle_path = os.path.join(Pyramid_Dir, k + '.p')
            with open(pickle_path, 'wb') as handle:
                pickle.dump(v, handle, protocol=4 )
                print(k + '\tit is Dumped into file:', pickle_path)
                print(k + '\tthe length of it is   :', v['length'])
        print('*'*40, '\n')

        ################################################################################
        Path_Key = os.path.join(Data_Dir, 'Vocab')
        if not os.path.exists(Path_Key):
            os.makedirs(Path_Key)

        for k, v in cls.VOCAB[Path_Key].items():
            
            pickle_path = os.path.join(Path_Key, k + '.voc')
            with open(pickle_path, 'wb') as handle:
                # v is (LGU, DGU)
                pickle.dump(v, handle, protocol=4)
                print(k + '\tis Dumped into file:', pickle_path)
                print(k + '\tthe length of it is   :', len(v[0]))
            
            channel_name_path = os.path.join(Path_Key, k + '.tsv')
            writeGrainList2File(channel_name_path, v[0])
            print('\t\tWrite to:', channel_name_path)

        ################################################################################ 
        pickle_path = os.path.join(Path_Key, 'token.voc')
        with open(pickle_path, 'wb') as handle:
            # v is (LGU, DGU)
            pickle.dump(cls.TokenVocab, handle, protocol=4)
            print('token   ' + '\tis Dumped into file:', pickle_path)
            print('token   ' + '\tthe length of it is   :', len(cls.TokenVocab[0]))
        
        channel_name_path = os.path.join(Path_Key, 'token.tsv')
        writeGrainList2File(channel_name_path, cls.TokenVocab[0])
        print('\t\tWrite to:', channel_name_path)

        pickle_path = os.path.join(Path_Key, 'token.freq')
        with open(pickle_path, 'wb') as handle:
            pickle.dump(cls.idx2freq, handle, protocol=4)
        print('*'*40)
        
    @classmethod
    def _load_tokenvocab_from_disk(cls):
        ################################################################################
        Path_Key = os.path.join(cls.Data_Dir, 'Vocab')
        pickle_path = os.path.join(Path_Key, 'token.freq')
        with open(pickle_path, 'rb') as handle:
            cls.idx2freq = pickle.load(handle)

        ################################################################################
        cls.VOCAB[Path_Key] = {} if Path_Key not in cls.VOCAB else cls.VOCAB[Path_Key]

        pickle_path = os.path.join(Path_Key, 'token.voc')
        with open(pickle_path, 'rb') as handle:
            cls.TokenVocab = pickle.load(handle)

        orig_vocab_length = len(cls.TokenVocab[0])
        cls.original_vocab_token_num = orig_vocab_length
        cls.current_vocab_token_num  = orig_vocab_length

    @classmethod
    def _buildGVforToken(cls, max_vocab_token_num = None, min_token_freq = 1):
        # both two inputs are valid
        try:
            if min_token_freq == cls.min_token_freq:
                return cls.TokenVocab 
        except:
            cls.min_token_freq = min_token_freq

        if min_token_freq < cls.min_token_freq:
            # reset all token vocab from disk
            cls._load_tokenvocab_from_disk()

        cls.idx2freq = cls.idx2freq[:max_vocab_token_num]
        LTU, DTU = cls.TokenVocab
        LTU = LTU[:max_vocab_token_num]
        DTU = {}
        for newidx in range(max_vocab_token_num):
            DTU[LTU[newidx]] = newidx
        cls.TokenVocab = LTU, DTU
        # print(max_vocab_token_num, min_token_freq)
        cls.current_vocab_token_num = max_vocab_token_num
        cls.min_token_freq = min_token_freq
        return cls.TokenVocab 

    @classmethod
    def INIT_FROM_PICKLE(cls, Data_Dir, min_token_freq = 1):
        assert os.path.exists(Data_Dir) 
        cls.Data_Dir = Data_Dir
        ################################################################################
        layer_names = ['CORPUS', 'GROUP','TEXT', 'SENT','TOKEN' ]
        for layer_name in layer_names:
            pickle_path = os.path.join(Data_Dir, 'Pyramid', layer_name + '.p')
            with open(pickle_path, 'rb') as handle:
                v = pickle.load(handle)
                setattr(cls, layer_name,  v)
                print(layer_name + '\tread from pickle file :', pickle_path)
                print(layer_name + '\tthe length of it is   :', v['length'])
        print('*'*40, '\n')

        cls.Channel_Hyper_Path = cls.TOKEN['Channel_Hyper_Path']
        # idx2freq and TokenVocab
        cls._load_tokenvocab_from_disk()
        max_vocab_token_num = get_num_freq(cls.idx2freq, min_token_freq = min_token_freq)
        cls._buildGVforToken(max_vocab_token_num, min_token_freq)

        folder = 'F' + str(min_token_freq) 
        Path_Key = os.path.join(cls.Data_Dir, 'Vocab', folder)
        if not os.path.exists(Path_Key):
            os.makedirs(Path_Key)

    @classmethod
    def _getGVfromVocab(cls, Path_Key, channel_name):
        return cls.VOCAB[Path_Key][channel_name]

    @classmethod
    def _getGVfromDisk(cls, Path_Key, channel_name):
        cls.VOCAB[Path_Key] = cls.VOCAB[Path_Key] if Path_Key in cls.VOCAB else {}
        channel_name_pickle = os.path.join(Path_Key, channel_name + '.voc')
        GrainVocab = readPickleFile2GrainUnique(channel_name_pickle)
        cls.VOCAB[Path_Key][channel_name] = GrainVocab
        return GrainVocab 

    @classmethod
    def _buildGVforHyper(cls, Path_Key, channel, channel_name):
        # build GV for hyper field
        print('\t\tBuild GrainUnique for channel:', channel_name)
        ############################################# Generate New LGU for CTX_Dep
        channel, Min_Ngram, Max_Ngram, end_grain, min_grain_freq, tagScheme = getChannelName(channel, channel_name = channel_name, style = 'extract')
        print(channel, Max_Ngram, end_grain, tagScheme)
        
        ch = 'annoE' if 'annoR' == channel else channel
        # this many fall into loops
        # BIOES_GU = cls.getGrainVocab(ch, tagScheme = 'BIOES') # cautions: must get the corresponding base GU.
        try:
            # generall, we hope this can give us results.
            BIOES_GU = cls._getGVfromVocab(Path_Key, ch + '-bioes')
        except:
            BIOES_GU = cls._getGVfromDisk(Path_Key, ch + '-bioes')

        BIOES_GU_neat = BIOES_GU[0]

        LGU_neat = list(set([trans_bioesTag(channel, i, tagScheme) for i in BIOES_GU_neat]))
        LGU_neat.sort()

        LGU = LGU_neat
        DGU = dict(zip(LGU, range(len(LGU))))

        ############################################# New LGU
        GrainVocab = (LGU, DGU)

        cls.VOCAB[Path_Key] = cls.VOCAB[Path_Key] if Path_Key in cls.VOCAB else {}
        cls.VOCAB[Path_Key][channel_name] = GrainVocab
        
        # how about this one?
        
        channel_name_pickle = os.path.join(Path_Key, channel_name + '.voc')
        with open(channel_name_pickle, 'wb') as handle:
            pickle.dump(GrainVocab, handle)
        print('\t\tWrite to:', channel_name_pickle)

        channel_name_path = os.path.join(Path_Key, channel_name + '.tsv')
        writeGrainList2File(channel_name_path, GrainVocab[0])
        print('\t\tWrite to:', channel_name_path)
        return GrainVocab

    @classmethod
    def _buildGVforSub(cls, Path_Key, channel, channel_name, min_token_freq = 1):

        print('\t\tBuild Grain Uniqe and LookUp Table for channel:', channel_name)

        channel, Min_Ngram, Max_Ngram, end_grain, min_grain_freq, tagScheme = getChannelName(channel, channel_name = channel_name, style = 'extract')
        
        # build GV and LKP
        GrainVocab, LKP, grainidx2freq = get_GU_or_LKP(cls.TokenVocab, cls.idx2freq, channel, Min_Ngram , Max_Ngram , end_grain , min_grain_freq )
        
        # save to VOCAB
        cls.VOCAB[Path_Key] = cls.VOCAB[Path_Key] if Path_Key in cls.VOCAB else {}
        cls.VOCAB[Path_Key][channel_name] = GrainVocab

        # save GV to disk
        channel_name_pickle = os.path.join(Path_Key, channel_name + '.voc')
        with open(channel_name_pickle, 'wb') as handle:
            pickle.dump(GrainVocab, handle)
        print('\t\tWrite to:', channel_name_pickle)
        
        channel_name_path = os.path.join(Path_Key, channel_name + '.tsv')
        writeGrainList2File(channel_name_path, GrainVocab[0])
        print('\t\tWrite to:', channel_name_path)

        # save LKP to LOOKUP
        cls.LOOKUP[Path_Key] = cls.LOOKUP[Path_Key] if Path_Key in cls.LOOKUP else {}
        cls.LOOKUP[Path_Key][channel_name] = LKP
        # save LKP to disk
        pickle_path = os.path.join(Path_Key,  channel_name + '.lkp')
        with open(pickle_path, 'wb') as handle:
            pickle.dump(LKP, handle)
        # assert len(LookUp) == len(cls.TokenVocab[0])
        print('\t\tWrite to:', pickle_path)

        # save LKP to LOOKUP
        cls.FREQ[Path_Key] = cls.FREQ[Path_Key] if Path_Key in cls.FREQ else {}
        cls.FREQ[Path_Key][channel_name] = grainidx2freq
        # save LKP to disk
        pickle_path = os.path.join(Path_Key,  channel_name + '.freq')
        with open(pickle_path, 'wb') as handle:
            pickle.dump(grainidx2freq, handle)
        print('\t\tWrite to:', pickle_path)
       
        return GrainVocab

    @classmethod
    def getGrainVocab(cls, channel, Min_Ngram = 1, Max_Ngram = 1, end_grain = False, tagScheme = 'BIO', min_grain_freq = 1, channel_name = None,
                      min_token_freq = None, Data_Dir = None,  **kwargs):

        # for token, only return itself token vocab
        if channel == 'token':
            if min_token_freq == cls.min_token_freq or not min_token_freq:
                return cls.TokenVocab
            else:
                raise('Error in getting token vocab, only one method is implemented now.')

        # find the Data_Dir
        Data_Dir = cls.Data_Dir if not Data_Dir else Data_Dir

        # build the path key and the channel name
        if not channel_name:
            channel_name = getChannelName(channel, Min_Ngram = Min_Ngram, Max_Ngram = Max_Ngram, end_grain = end_grain, 
                                          min_grain_freq = min_grain_freq, tagScheme = tagScheme)
        else:
            channel, Min_Ngram, Max_Ngram, end_grain, min_grain_freq, tagScheme = getChannelName(channel = channel, channel_name = channel_name, style = 'extract')

        # build the Path_Key
        if channel not in cls.CONTEXT_IND_CHANNELS:
            Path_Key = os.path.join(Data_Dir, 'Vocab')
        else:
            min_token_freq = cls.min_token_freq if not min_token_freq else min_token_freq
            # it cannot makes a folder
            folder = 'F' + str(min_token_freq) 
            Path_Key = os.path.join(Data_Dir, 'Vocab', folder)

        # try to return the GV
        try:
            # generall, we hope this can give us results.
            return cls._getGVfromVocab(Path_Key, channel_name)
        except:
            try:
                return cls._getGVfromDisk(Path_Key, channel_name)
            except:
                # if it is the cls.Data_Dir, we can build GV; else we cannot.
                if cls.Data_Dir == Data_Dir:
                    # to check whether the channel is hyper or sub
                    if channel not in cls.CONTEXT_IND_CHANNELS:
                        # also include token
                        return cls._buildGVforHyper(Path_Key, channel, channel_name)
                    else:
                        # TODO incorporate max grain num and min grain freq into channel_name in the future
                        return cls._buildGVforSub(Path_Key, channel, channel_name, min_token_freq = min_token_freq)
                else:
                    channel_name_pickle = os.path.join(Path_Key, channel_name + '.voc')
                    print('In', Data_Dir, 'there is no GrainUnqiue for:', channel_name)
                    print('Error in:', channel_name_pickle)

    @classmethod
    def getLookUp(cls, channel = None, Min_Ngram = 1, Max_Ngram = 1, end_grain = False, min_grain_freq = 1, channel_name = None,
                  Data_Dir = None,  min_token_freq = None, **kwargs):

        # find the Data_Dir
        Data_Dir = cls.Data_Dir if not Data_Dir else Data_Dir

        if not channel_name:
            channel_name = getChannelName(channel, Min_Ngram = Min_Ngram, Max_Ngram = Max_Ngram, end_grain = end_grain, 
                                          min_grain_freq = min_grain_freq)
        else:
            channel, Min_Ngram, Max_Ngram, end_grain, min_grain_freq, tagScheme = getChannelName(channel = channel, channel_name = channel_name, style = 'extract')

        # build the Path_Key
        min_token_freq = cls.min_token_freq if not min_token_freq else min_token_freq
        # it cannot makes a folder
        folder = 'F' + str(min_token_freq) 
        Path_Key = os.path.join(Data_Dir, 'Vocab', folder)

        try:
            # generall, we hope this can give us results.
            LKP = cls.LOOKUP[Path_Key][channel_name]
            TV = cls.getGrainVocab(channel = 'token', Data_Dir = Data_Dir)
            return LKP, TV
        except:
            try:
                path = os.path.join(Path_Key, channel_name + '.lkp')
                # assert os.path.isfile(path)
                with open(path, 'rb') as handle:
                    LKP = pickle.load(handle)
                
                cls.LOOKUP[Path_Key] = cls.LOOKUP[Path_Key] if Path_Key in cls.LOOKUP else {}
                cls.LOOKUP[Path_Key][channel_name] = LKP

                # read from other pyramid is still no valid
                TV = cls.getGrainVocab(channel = 'token', Data_Dir = Data_Dir)
                print('Get LookUp Table for Channel:', channel_name)
                return LKP, TV
            except:
                print('\tIn', Path_Key, 'there is no LookUp Table for:', channel_name)
                print('\tError in:', path)
                return None

    @classmethod
    def getFreq(cls, channel = None, Min_Ngram = 1, Max_Ngram = 1, end_grain = False, channel_name = None,
                  Data_Dir = None, min_grain_freq = 1, min_token_freq = None, **kwargs):

        # find the Data_Dir
        Data_Dir = cls.Data_Dir if not Data_Dir else Data_Dir
        
        if not channel_name:
            channel_name = getChannelName(channel, Min_Ngram = Min_Ngram, Max_Ngram = Max_Ngram, end_grain = end_grain, 
                                          min_grain_freq = min_grain_freq)
        else:
            channel, Min_Ngram, Max_Ngram, end_grain, min_grain_freq, tagScheme = getChannelName(channel = channel, channel_name = channel_name, style = 'extract')

        # build the Path_Key
        min_token_freq = cls.min_token_freq if not min_token_freq else min_token_freq
        # it cannot makes a folder
        folder = 'F' + str(min_token_freq) 
        Path_Key = os.path.join(Data_Dir, 'Vocab', folder)

        try:
            # generall, we hope this can give us results.
            Freq = cls.FREQ[Path_Key][channel_name]
            return Freq
        except:
            try:
                path = os.path.join(Path_Key, channel_name + '.freq')
                # assert os.path.isfile(path)
                with open(path, 'rb') as handle:
                    Freq = pickle.load(handle)
                
                cls.FREQ[Path_Key] = cls.FREQ[Path_Key] if Path_Key in cls.FREQ else {}
                cls.FREQ[Path_Key][channel_name] = Freq
                # read from other pyramid is still no valid
                return Freq
            except:
                print('\tIn', Path_Key, 'there is no LookUp Table for:', channel_name)
                print('\tError in:', path)
                return None


    @classmethod
    def getTrans(cls, channel, tagScheme, Data_Dir = None, GU = None):
        # get Information for TRANS
        channel_name = getChannelName(channel, tagScheme = tagScheme)
        Data_Dir = cls.Data_Dir if not Data_Dir else Data_Dir
        
        if not GU: 
            Path_Key = Data_Dir
            GU = cls.getGrainVocab(channel, tagScheme = tagScheme)
        else:
            Path_Key = Data_Dir + '-tmp'

        try:
            return cls.TRANS[Path_Key][channel_name] # be caution, this may occur an issue
        except:
            ch = 'annoE' if 'anno' in  channel else channel
            BIOES_GU = cls.getGrainVocab(ch, tagScheme = "BIOES")
            new_GU   = GU
            BIOES_LGU= BIOES_GU[0]
            new_DGU  = new_GU[1]
            cls.TRANS[Path_Key] = {} if Path_Key not in cls.TRANS else cls.TRANS[Path_Key]
            cls.TRANS[Path_Key][channel_name] = {str(idx): new_DGU[ trans_bioesTag(channel, bioesTag, tagScheme)] 
                                                 for idx, bioesTag in enumerate(BIOES_LGU)}
            return cls.TRANS[Path_Key][channel_name] 

    @classmethod
    def BUILD_GV_LKP(cls, CHANNEL_SETTINGS_TEMPLATE = None):

        cls.CHANNEL_SETTINGS, cls.channels_folderName = get_Channel_Settings(CHANNEL_SETTINGS_TEMPLATE)

        for channel in cls.CHANNEL_SETTINGS:
            print('Deal with the Channel:', channel)
            channel_setting = cls.CHANNEL_SETTINGS[channel]
            Min_Ngram    = channel_setting.get('Min_Ngram', 1)
            Max_Ngram    = channel_setting.get('Max_Ngram', 1)
            end_grain    = channel_setting.get('end_grain', False)
            tagScheme    = channel_setting.get('tagScheme', 'BIO')
            min_grain_freq = channel_setting.get('min_grain_freq', 1)
            print('Current Channel is       ', '\t', channel)
            print('Current Channel Max_Ngram', '\t', Max_Ngram)
            cls.getGrainVocab(channel, Min_Ngram = Min_Ngram, Max_Ngram = Max_Ngram, end_grain = end_grain, tagScheme = tagScheme, min_grain_freq = min_grain_freq)
    

def convert_Char_2_Word_BasicObject(BasicObject, use_channel = 'pos', MaxTokenUnique = False):
    
    CORPUS = BasicObject.CORPUS.copy()
    SENT = BasicObject.SENT.copy()
    SENT['EndIDXTokens'] = []
    TOKENLevel = 'word'
    TOKEN = {}
    TOKEN['TOKENLevel']    = TOKENLevel
    CORPUSPath = CORPUS['CORPUSPath']
    ch = use_channel
    posTokenIndex = BasicObject.TOKEN[ch + 'TokenIndex']
    posGU = BasicObject.getGrainUnique(ch, tagScheme='BIOES')
    posLGU, posDGU = posGU

    if 'ANNOTokenIndex' in BasicObject.TOKEN:
        annoTokenIndex = BasicObject.TOKEN['ANNOTokenIndex']
        annoGU = BasicObject.getGrainUnique('annoE', tagScheme='BIOES')
        annoLGU, annoDGU = annoGU

    pos_all  = []
    anno_all = []
    word_all = []
    
    from .sentence import Sentence
    for i in range(BasicObject.SENT['length']):
        st = Sentence(i)
        sentence = st.sentence
        s, e = st.IdxTokenStartEnd
        pos_sent_index = posTokenIndex[s:e]
        
        pos_sent  = [posLGU[i] for i in pos_sent_index]

        if 'ANNOTokenIndex' in BasicObject.TOKEN:
            anno_sent_index = annoTokenIndex[s:e]
            anno_sent = [annoLGU[i] for i in anno_sent_index]

        if 'ANNOTokenIndex' in BasicObject.TOKEN:
            pos_sent_BIO = modify_wordBoundary_with_hyperBoundary(pos_sent, anno_sent) # convert to BIO
        else:
            pos_sent_BIO = modify_wordBoundary_with_hyperBoundary(pos_sent, pos_sent) # convert to BIO

        new_pos_seq = getSSET_from_CIT(pos_sent_BIO, pos_sent_BIO, join_char='*')
        pos_seq = [posDGU[i[-1]+'-S'] for i in new_pos_seq] # still pos-es
        pos_all.extend(pos_seq)

        if 'ANNOTokenIndex' in BasicObject.TOKEN:
            new_anno_seq = getSSET_from_CIT(anno_sent, pos_sent_BIO, join_char='*')
            anno_seq = [annoDGU[trans_charLabels_to_wordLabels(i[0], join_char = '*')] for i in new_anno_seq] # still annoE-es
            anno_all.extend(anno_seq)

        new_st_seq = getSSET_from_CIT(sentence, pos_sent_BIO, join_char='')
        st_seq = [i[0] for i in new_st_seq]
        word_all.extend(st_seq)

        assert len(st_seq) == len(pos_seq) 
        if 'ANNOTokenIndex' in BasicObject.TOKEN:
            assert len(st_seq) == len(anno_seq)
        lenSent = len(st_seq)
        try:
            SENT['EndIDXTokens'].append(SENT['EndIDXTokens'][-1] + lenSent)
        except:
            SENT['EndIDXTokens'].append(lenSent)
            

    TOKEN['ORIGToken']     = word_all
    TOKEN[ch+'TokenIndex'] = np.array(pos_all)

    if 'ANNOTokenIndex' in BasicObject.TOKEN:
        TOKEN['ANNOTokenIndex']= np.array(anno_all)

    TOKEN['length']        = len(word_all)

    SENT['EndIDXTokens']    = np.array(SENT['EndIDXTokens'])
    SENT['length']          = len(SENT['EndIDXTokens'])

    assert SENT['length'] == BasicObject.SENT['length']

    ########################################################
    print('Total Num of All    Tokens', len(TOKEN['ORIGToken']))
    # TOKEN['ORIGToken'] = list(TOKEN['ORIGToken'])
    # print(datetime.now())
    TOKEN['ORIGTokenIndex'], LTU, DTU =  buildTokens(TOKEN['ORIGToken'], MaxTokenUnique)
    # print(datetime.now())
    del TOKEN['ORIGToken']

    print('Total Num of Unique Tokens', len(LTU))
    corpus_name = CORPUSPath.split('/')[-2]
    TokenNum_Dir = os.path.join('data', corpus_name, TOKENLevel, 'Token'+str(len(LTU)))
    # print(TokenNum_Dir)
    CORPUS['TokenNum_Dir'] = TokenNum_Dir
    if not os.path.exists(TokenNum_Dir):
        os.makedirs(TokenNum_Dir)

    # token GrainUnique (TokenUnique)
    GRAIN_UNI = {}
    GRAIN_UNI[TokenNum_Dir] = {'token': (LTU, DTU)}
    channel_name_path = os.path.join(TokenNum_Dir, 'token.tsv')
    writeGrainList2File(channel_name_path, LTU)
    print('\t\tWrite to:', channel_name_path)

    ########################################################
    if 'ANNOTokenIndex' in BasicObject.TOKEN:
        channel_name = 'annoE-bioes'
        GRAIN_UNI[TokenNum_Dir][channel_name] = annoGU
        LGU = annoGU[0]
        channel_name_path = os.path.join(TokenNum_Dir, channel_name+ '.tsv')
        writeGrainList2File(channel_name_path, LGU)
        print('\t\tWrite to:', channel_name_path)
        # print(LGU)

    ########################################################
    channel_name = 'pos-bioes'
    GRAIN_UNI[TokenNum_Dir][channel_name] = posGU
    LGU = posGU[0]
    channel_name_path = os.path.join(TokenNum_Dir, channel_name+ '.tsv')
    writeGrainList2File(channel_name_path, LGU)
    print('\t\tWrite to:', channel_name_path, '\n')
    # print(LGU)

    BasicObject.CORPUS= CORPUS
    BasicObject.SENT  = SENT
    BasicObject.TOKEN = TOKEN
    BasicObject.TokenUnique  = (LTU, LGU)
    BasicObject.GRAIN_UNI    = GRAIN_UNI
    BasicObject.TokenNum_Dir = TokenNum_Dir
    
    BasicObject.OBJECT_TO_PICKLE()
    
    return BasicObject



########################## text.py #######################
import re
from bisect import bisect

from .base import BasicObject
# from .corpus import Corpus
# from .folder import Folder 
# from .text import Text
from .sentence import Sentence
from .token import Token

from .utils.pyramid import segText2Sents, get_lines_with_position, read_file_chunk_string

class Text(BasicObject):

    _text = None

    def __init__(self, Idx = None, text = None):
        self.Idx = Idx
        if text:
            self._text = text

    @property
    def name(self):
        return self.TEXT['ORIGFileName'][self.Idx] if 'ORIGFileName' in self.TEXT else 'TxT' 

    @property
    def text(self):
        if self._text:
            return self._text
        else:
            start_position, end_position = self.start_end_position('token')
            # s, e = self.IdxSentStartEnd
            # num_lines = e - s 
            return read_file_chunk_string(self.Channel_Hyper_Path['token'], start_position, end_position)

    def get_stored_hyper(self, channel):
        start_position, end_position = self.start_end_position(channel)
        return read_file_chunk_string(self.Channel_Hyper_Path[channel], start_position, end_position)

    def get_stored_hypertagscheme(self, channel, tagScheme):
        # here channel should exclude token
        grain_idx = re.split(' |\n', self.get_stored_hyper(channel))
        bioes2tag = self.getTrans(channel, tagScheme)
        # GV = self.getGrainVocab(channel, tagScheme)
        # shall we check its insanity?
        return [bioes2tag[vocidx] for vocidx in grain_idx]

    def get_stored_hyperstring(self, channel, tagScheme):
        vocidx2grain = self.getGrainVocab(channel, tagScheme = tagScheme)[0]
        grain_idx = self.get_stored_hypertagscheme(channel, tagScheme)
        return [vocidx2grain[vocidx] for vocidx in grain_idx]

    def start_end_position(self, channel):
        startsentIdx, endsentIdx = self.IdxSentStartEnd
        start_position = self.SENT[self.Channel_Hyper_Path[channel]][startsentIdx-1] if startsentIdx != 0 else 0
        end_position   = self.SENT[self.Channel_Hyper_Path[channel]][endsentIdx - 1]
        return start_position, end_position

    @property 
    def IdxCorpus(self):
        return bisect(self.CORPUS['EndIDXFolders'] , self.IdxFolder)

    @property
    def IdxFolder(self):
        return bisect(self.GROUP['EndIDXTexts'] , self.Idx)

    @property
    def IdxSentStartEnd(self):
        s, e = self.Idx, self.Idx+1
        s = self.TEXT['EndIDXSents'][s-1] if s != 0 else 0
        e = self.TEXT['EndIDXSents'][e-1]
        return s, e 

    @property
    def IdxTokenStartEnd(self):
        s, e = self.IdxSentStartEnd
        s = self.SENT['EndIDXTokens'][s-1] if s != 0 else 0
        e = self.SENT['EndIDXTokens'][e-1]
        return s, e 

    @property
    def Corpus(self):
        from .corpus import Corpus 
        return Corpus(self.IdxCorpus)

    @property
    def Folder(self):
        from .folder import Folder 
        return Folder(self.IdxFolder)

    @property
    def Sentences(self):
        if self._text:
            text = segText2Sents(self._text, method = self.TEXT['Text2SentMethod'])
            return [Sentence(sentence = sent) for sent in text]
        else:
            return [Sentence(Idx) for Idx in range(*self.IdxSentStartEnd)]

    @property
    def Tokens(self):
        if self._text:
            return sum([st.Tokens for st in self.Sentences], [])
        else:
            return [Token(Idx)    for Idx in range(*self.IdxTokenStartEnd)]

    @property
    def length(self):
        if self._text:
            length = len(self.Tokens)
        else:
            s, e = self.IdxTokenStartEnd 
            length = e - s
        return length
    
    def __repr__(self):
        ctx =  self.name + " " + str(self.Idx)  if type(self.Idx) == int else 'New'
        return "<txt " + ctx + ">"




import os
import time
import codecs
import optparse 
import pickle

import numpy as np
import pandas as pd

from .crftools import get_sent_strfeats, crf_test
from .evals import read_target_seq, extractSET
from .nlptext_lite.sentence import Sentence





def medpos_tagger(sent, 
                  model = None, 
                  Channel_Settings = None, 
                  get_seq = False, 
                  get_entities = True, 
                  name = '1'):
    
    '''
        basically from crf_test
        sent: a sentence, could be without annotation
    '''

    current_dir_old = os.path.abspath(os.path.dirname(__file__))
    current_dir, _ = os.path.split(current_dir_old)
    model = 'model/MedPos/char/T_b-n1t1-f1_m-n1t1-f1_r-n1t1-f1_P-bio_E-bio'
    model = os.path.join(current_dir, model)
    # print(model)

    sent = Sentence(sentence=[i.replace(' ', '@') for i in sent], tokenLevel= 'char')
    # model = model + '_' + str(cross_idx)
    if not Channel_Settings:
        with open(model + '/para.p', 'rb') as f:
            Channel_Settings = pickle.load(f)
    # 1. get sentence feats
    # hopefully, the model_config is included in model_path
    # if not os.path.exists():

    feats_data_path   = os.path.join(current_dir, name + '_tagger_feats.txt') 
    results_data_path = os.path.join(current_dir, name + '_tagger_results.txt') 
    model_path = model + '/model'
    # print(feats_data_path)

    # get Channel_Settings
    # get use sent strfeats or sent vecfeats settings
    df = get_sent_strfeats(sent, Channel_Settings, train = False)
    # print(df)
    df.to_csv(feats_data_path, sep = '\t', encoding = 'utf=8', header = False, index = False )
    # 2. save the sentence feats to a file
    
    # 3. tag a sentence
    crf_test(feats_data_path, model_path, results_data_path)

    # 4. read and parse the result to pred_SSET
    # get a tag_seq
    # list of tuple (score, result)

    tag_seq  = read_target_seq(results_data_path)
    if get_seq: 
        return tag_seq
    else:
        pred_SET = extractSET(tag_seq)
        return pred_SET




import os
from bisect import bisect
import numpy as np 

from .base import BasicObject
from .utils.channel import getChannelGrain4Token, getChannelName
from .utils.infrastructure import UNK_ID

class Token(BasicObject):

    _token = None
    _Sentence = None 
    _IdxTokenInSent = None

    def __init__(self, Idx = None, token = None, Sentence = None, IdxTokenInSent = None):
        self.Idx = Idx
        if token:
            self._token = token
        if Sentence:
            self._Sentence = Sentence # type is nlptext.Sentence
            self._IdxTokenInSent = IdxTokenInSent

    # index: is the index in 
    # @property
    # def idx(self):
    #     if self._token:
    #         return self.TokenVocab[1].get(self._token, len(TokenVocab[1]))
    #     else:
    #         # remove this out 
    #         start_position = self.start_position('token')
    #         return self.TOKEN['ORIGTokenIndex'][self.Idx]

    # @property
    # def token(self):
    #     if self._token:
    #         return self._token
    #     else:
    #         # there is another quick method
    #         # logic: get the sentence first, then get the token
    #         return self.TokenUnique[0][self.index] # quick? or slow?

    def start_position(self, channel):
        # the byte idx where this token's sentence starts from.
        # Idx = self.IdxText
        start_position = self.SENT[self.Channel_Hyper_Path[channel]][self.IdxText-1] if self.IdxText != 0 else 0
        return start_position

    @property 
    def IdxCorpus(self):
        return bisect(self.CORPUS['EndIDXFolders'] , self.IdxFolder)

    @property
    def IdxFolder(self):
        return bisect(self.GROUP['EndIDXTexts'] , self.IdxText)

    @property
    def IdxText(self):
        return bisect(self.TEXT['EndIDXSents'] , self.IdxSent)

    @property
    def IdxSent(self):
        return bisect(self.SENT['EndIDXTokens'] , self.Idx)

    @property
    def IdxTokenInSent(self):
        if type(self._IdxTokenInSent) == int:
            return self._IdxTokenInSent
        else:
            return self.Idx - self.Sentence.IdxTokenStartEnd[0]
    
    @property
    def Corpus(self):
        from .corpus import Corpus 
        return Corpus(self.IdxCorpus)

    @property
    def Folder(self):
        from .folder import Folder 
        return Folder(self.IdxFolder)

    @property
    def Text(self):
        from .text import Text 
        return Text(self.IdxText)

    @property
    def Sentence(self):
        if self._Sentence:
            return self._Sentence
        else:
            from .sentence import Sentence
            return Sentence(self.IdxSent)

    def __repr__(self):
        return "<tk "  + self.token + " >"

    




##################################################################################################TEXT-ANNO
def getCITText(strText, SSETText, TOKENLevel='char'):
    # assert len(SSETText) > 0 
    if TOKENLevel == 'char':
        for sset in SSETText:
            try:
                assert strText[sset[1]: sset[2]] == sset[0]
            except:
                print('strText:', strText[sset[1] : sset[2]])
                print('SSETText:', sset[0])
        CITAnnoText = []
        for sset in SSETText:
            # BIOES
            strAnno, s, e, tag = sset
            CIT = [[c, s + idx, tag+ '-I']  for idx, c in enumerate(strAnno)]
            CIT[-1][2] = tag + '-E'
            CIT[ 0][2] = tag + '-B'
            if len(CIT) == 1:
                CIT[0][2] = tag + '-S' 
            CITAnnoText.extend(CIT)

        # print(strAnnoText)
        CITText = [[char, idx, 'O'] for idx, char in enumerate(strText)]
        for citAnno in CITAnnoText:
            c, idx, t = citAnno
            assert CITText[idx][0] == c
            CITText[idx] = citAnno

    elif TOKENLevel == 'word':
        strText = strText.split(' ')
        for sset in SSETText:
            try:
                assert ' '.join(strText[sset[1]: sset[2]]) == sset[0]
            except:
                print('strText:', strText[sset[1] : sset[2]])
                print('SSETText:', sset[0])

        CITAnnoText = []
        for sset in SSETText:
            # BIOES
            strAnno, s, e, tag = sset
            # this is important
            strAnno = strAnno.split(' ')
            CIT = [[c, s + idx, tag+ '-I']  for idx, c in enumerate(strAnno)]
            CIT[-1][2] = tag + '-E'
            CIT[ 0][2] = tag + '-B'
            if len(CIT) == 1:
                CIT[0][2] = tag + '-S' 
            CITAnnoText.extend(CIT)

        # print('\nCITAnnoText\n')
        # print(CITAnnoText)
        # print('\n')
        # print(strText)
        # print('\n')
        CITText = [[char, idx, 'O'] for idx, char in enumerate(strText)]
        for citAnno in CITAnnoText:
            c, idx, t = citAnno
            assert CITText[idx][0] == c
            CITText[idx] = citAnno
        # print('                 English CITText')
        # print(CITText)
    return CITText

def getCITSents(tokenizedSents, CITText, TOKENLevel='char'):
    # if TOKENLevel == 'char':

    # if TOKENLevel == 'word':
    #     strSents = [tokenizedSent.split(' ') for tokenizedSent in tokenizedSents]
    # print('\nin strSents\n')
    # print(strSents)
    # print(CITText)
    lenLastSent = 0
    collapse    = 0 # don't need to move 
    CITSents = []
    for tokenizedSent in tokenizedSents:

        CITSent = []
        for sentTokenIdx, c in enumerate(tokenizedSent):
            # sentTokenIdx = txtTokenIdx - lenLastSent - collapse
            txtTokenIdx = sentTokenIdx + lenLastSent + collapse
            cT, _, tT = CITText[txtTokenIdx]
            while c != cT and c != ' ':
                # print(c, cT)
                collapse = collapse + 1
                txtTokenIdx = sentTokenIdx + lenLastSent + collapse
                if txtTokenIdx >= len(CITText):
                    raise ValueError('You cannot find a good token!')
                cT, _, tT = CITText[txtTokenIdx]
            CITSent.append([c, sentTokenIdx, tT])
        lenLastSent = lenLastSent + len(tokenizedSent)
        CITSents.append(CITSent)
    # CITSents
    # Here we get CITSents  
    return CITSents
       
def getSSET_from_CIT(orig_seq, tag_seq, tag_seq_tagScheme = 'BIO', join_char = ''):
    # orig_seq is sentence without start or end
    # tag_seq may have start or end
    if tag_seq[0] == '</start>':
        tag_seq = tag_seq[1:-1]
        
    tagScheme = tag_seq_tagScheme
    if tagScheme == 'BIOES':
        tag_seq = [i.replace('-S', '-B').replace('-E', '-I') for i in tag_seq]
    elif tagScheme == 'BIOE':
        tag_seq = [i.replace('-E', '-I') for i in tag_seq]
    elif tagScheme == 'BIOS':
        tag_seq = [i.replace('-S', '-B') for i in tag_seq]
    elif tagScheme == 'BIO':
        pass
    else:
        print('The tagScheme', tagScheme, 'is not supported yet...')
    
    # use BIO tagScheme
    CIT = list(zip(orig_seq, range(len(orig_seq)), tag_seq))
    taggedCIT = [cit for cit in CIT if cit[2]!= 'O']
    
    startIdx = [idx for idx in range(len(taggedCIT)) if taggedCIT[idx][2][-2:] == '-B']
    startIdx.append(len(taggedCIT))

    entitiesList = []
    for i in range(len(startIdx)-1):
        entityAtom = taggedCIT[startIdx[i]: startIdx[i+1]]
        string = join_char.join([cit[0] for cit in entityAtom])
        start, end = entityAtom[0][1], entityAtom[-1][1] + 1
        tag = entityAtom[0][2].split('-')[0]
        entitiesList.append((string, start, end, tag))

    # if join_char == '*':
    #     a = [set([t.split('-')[0] for t in i[0].split('*')]) for i in entitiesList]
    #     for i in a:
    #         if len(i) > 1:
    #             pprint(list(zip(orig_seq, tag_seq)))
                
    return entitiesList


from .base import BasicObject
# from .corpus import Corpus
# from .folder import Folder 
from .text import Text
from .sentence import Sentence
from .token import Token

from bisect import bisect


class Folder(BasicObject):
    def __init__(self, Idx = 0):
        self.Idx = Idx

    @property
    def name(self):
        return self.GROUP['group_names'][self.Idx]

    @property 
    def IdxCorpus(self):
        return bisect(self.CORPUS['EndIDXFolders'] , self.Idx)


    @property
    def IdxTextStartEnd(self):
        s, e = self.Idx, self.Idx + 1
        s = self.GROUP['EndIDXTexts'][s-1] if s != 0 else 0
        e = self.GROUP['EndIDXTexts'][e-1]
        return s, e 

    @property
    def IdxSentStartEnd(self):
        s, e = self.IdxTextStartEnd
        s = self.TEXT['EndIDXSents'][s-1]   if s != 0 else 0
        e = self.TEXT['EndIDXSents'][e-1]
        return s, e 

    @property
    def IdxTokenStartEnd(self):
        s, e = self.IdxSentStartEnd
        s = self.SENT['EndIDXTokens'][s-1]  if s != 0 else 0
        e = self.SENT['EndIDXTokens'][e-1]
        return s, e 

    @property
    def Corpus(self):
        from .corpus import Corpus 
        return Corpus(self.IdxCorpus)
    
    @property
    def Texts(self):
        return [Text(Idx)     for Idx in range(*self.IdxTextStartEnd)]

    @property
    def Sentences(self):
        return [Sentence(Idx) for Idx in range(*self.IdxSentStartEnd)]

    # this seems to handle
    @property
    def Tokens(self):
        return [Token(Idx)    for Idx in range(*self.IdxTokenStartEnd)]


    @property
    def TokensStartEnd(self):
        return sum([st.TokensStartEnd for st in self.Sentences], [])

    def __repr__(self):
        return "<gr " + self.name + ' ' + str(self.Idx) + ">"



import random
import time
import numpy as np

def get_train_test_valid(total_sent_num, seed = 10, prop = 0.8):
    np.random.seed(seed)
    train_num = int(total_sent_num*prop)
    test_num  = int((total_sent_num - train_num) /2)
    valid_num = total_sent_num - train_num - test_num
    
    total_sent_idx = list(range(total_sent_num))

    np.random.shuffle(total_sent_idx)

    train_sent_idx = total_sent_idx[:train_num]
    print('The num of train sentences:', len(train_sent_idx))

    test_sent_idx = total_sent_idx[train_num:train_num + test_num]
    print('The num of test  sentences:', len(test_sent_idx))

    valid_sent_idx = total_sent_idx[train_num + test_num:]
    print('The num of valid sentences:', len(valid_sent_idx))
    
    return train_sent_idx, test_sent_idx, valid_sent_idx


def getCrossValidationIdx(length, cross_num, pad = -1):
    
    length_pad = (int(length/cross_num) + 1 ) * cross_num
    #print(length_pad)
    new_list = list(range(length)) + [pad] * (length_pad - length)
    #print(new_list)

    cross_index = np.array(random.sample(new_list, length_pad))
    #cross_index.sort()
    return cross_index.reshape((cross_num, -1))


def loadData(sents, cross_num, seed = 10, cross_validation = False, cross_idx = 0, pad = -1):
    random.seed(seed)
    if not cross_validation:
        test_index = random.sample(range(len(sents)), int(len(sents)/cross_num))
        test_index.sort()
        testSents  = [sents[idx] for idx in test_index]
        trainSents = [sents[idx] for idx in range(len(sents)) if idx not in test_index]
        return trainSents, testSents
    else:
        cross_index = getCrossValidationIdx(len(sents), cross_num, pad = pad)
        test_index  = cross_index[cross_idx, ]
        testSents   = [sents[idx] for idx in test_index]
        trainSents  = [sents[idx] for idx in range(len(sents)) if idx not in test_index]
        return trainSents, testSents




import pickle

############### PART SpecialTokens

PAD   = '</pad>'
START = '</start>'
END   = '</end>'
UNK   = '</unk>'
specialTokens     = [ PAD, START, END, UNK]
specialTokensDict = {PAD: 0, START: 1, END: 2, UNK : 3, }
START_ID = specialTokensDict[START]
END_ID   = specialTokensDict[END]
UNK_ID   = specialTokensDict[UNK]


############### PART CODE
def any2unicode(text, encoding='utf8', errors='strict'):
    if isinstance(text, str):
        return text
    return str(text, encoding, errors=errors)

def strQ2B(ustring):
    rstring = ''
    for uchar in ustring:
        inside_code = ord(uchar)
        if inside_code == 12288:
            inside_code = 32
        elif (inside_code >= 65281 and inside_code <= 65374):
            inside_code -= 65248
        rstring += chr(inside_code)
    return rstring
############### PART Input and Output 

def fileReader(path):
    with open(path, 'r', encoding = 'utf-8') as f:
        text = f.read()
    return strQ2B(text)

def writeGrainList2File(channel_name_path, ListGrainUnique):
    with open(channel_name_path, 'w', encoding = 'utf-8') as f:
        for i in ListGrainUnique:
            i = '\\n' if i == '\n' else i
            f.write(i + '\n')

def readFile2GrainList(channel_name_path):
    ListGrainUnique = []
    with open(channel_name_path, 'r', encoding = 'utf-8') as f:
        for gr in f.readlines():
            gr = '\n' if  '\\n' in gr[:-1] else gr[:-1]
            ListGrainUnique.append(gr)
    return ListGrainUnique

def readPickleFile2GrainUnique(channel_name_path):
    with open(channel_name_path, 'rb') as handle:
        v = pickle.load(handle)
    return v  #(LGU, DGU)

############### PART BIOES Labels and Transformation.
# labels
# input: ['不确定', '修饰',  '检查', '疾病', '症状'] 
# tags
# ouput: ['O', '不确定-B', '不确定-I', '修饰-B', '修饰-I', '检查-B', '检查-I', '疾病-B', '疾病-I', '症状-B', '症状-I']
def getTagDict(labels, tagScheme = 'BIOES'):
    L = []
    suffices = ['-B', '-I']
    if 'O' in tagScheme:
        pref = ['O'] # remove speicial token
    else:
        pref = []
    if 'E' in tagScheme:
        suffices = suffices + ['-E']
    if 'S' in tagScheme:
        suffices = suffices + ['-S']
    tags = [i for i in labels if i !='O']
    for tag in tags:
        L.extend([tag+suff for suff in suffices])
    L.sort()
    L = pref + L
    return L

def trans_bioesTag(channel, bioesTag, tagScheme):

    if bioesTag in specialTokens:
        return bioesTag

    if 'S' not in tagScheme and 'E' not in tagScheme:
        i = bioesTag.replace('-S', '-B').replace('-E', '-I')
    elif 'S' not in tagScheme:
        i = bioesTag.replace('-S', '-B')
    elif 'E' not in tagScheme:
        i = bioesTag.replace('-E', '-I')
    else:
        i = bioesTag

    if channel == 'annoR':
        return i.split('-')[-1] 
    else:
        return i
        
############### PART Char To Word
def modify_wordBoundary_with_hyperBoundary(pos_sent, anno_sent):
    # return the new word_boundary with BIO tagScheme.
    pos_sent = [i.replace('-S', '-B').replace('-E', '-I') for i in pos_sent]
    
    for idx, tag in enumerate(anno_sent):
        if ('-B' == anno_sent[idx][-2:]  or '-S' == anno_sent[idx][-2:]):
            if pos_sent[idx][-2:] != '-B':
                # pprint([anno_sent[idx], pos_sent[idx]])
                pos_sent[idx] = pos_sent[idx].split('-')[0] + '-B'

        if ('-E' == anno_sent[idx][-2:]  or '-S' == anno_sent[idx][-2:]) and idx + 1 < len(pos_sent):
            if pos_sent[idx+1][-2:] != '-B':
                # pprint([anno_sent[idx], pos_sent[idx], pos_sent[idx+1]])
                # pprint(list(zip(anno_sent, pos_sent)))
                pos_sent[idx+1] = pos_sent[idx+1].split('-')[0] + '-B'
                
    return pos_sent

def trans_charLabels_to_wordLabels(string, join_char = '*'):
    labels = string.split('*')
    if ['O'] == list(set(labels)):
        return 'O'
    else:
        labels = [i.split('-') for i in labels]
        tag = list(set([i[0] for i in labels]))
        if len(tag) > 1:
            return tag, loc_indicator
        else:
            tag = tag[0]
            loc_indicator = [i[1] for i in labels]
            # print(string)
            if 'B'== loc_indicator[0] and 'E' == loc_indicator[-1]:
                # print(loc_indicator, '-->', 'S')
                return tag + '-S'
            elif ['S']  == loc_indicator:
                # print(loc_indicator, '-->', 'S')
                return tag + '-S'
            elif 'B' == loc_indicator[0]:
                # print(loc_indicator, '-->', 'B')
                return tag + '-B'
            elif 'E' == loc_indicator[-1]:
                # print(loc_indicator, '-->', 'E')
                return tag + '-E'
            elif ['I'] == list(set(loc_indicator)):
                # print(loc_indicator, '-->', 'I')
                return tag + '-I'
            else:
                # print('Jie -->', tag, loc_indicator)
                return tag, loc_indicator

############### PART OTHERS

def extractSET(tag_seq, exist_SE = False):
    '''
        SET: start, end, tag
        tag_seq: the hyper field sequence for this sentence
    '''
    if exist_SE:
        tag_seq = tag_seq[1:-1]

    IT = list(zip(range(len(tag_seq)), tag_seq))
    taggedIT = [it for it in IT if it[1]!= 'O']
    
    startIdx = [idx for idx in range(len(taggedIT)) if taggedIT[idx][1][-2:] == '-B' or taggedIT[idx][1][-2:] == '-S']
    startIdx.append(len(taggedIT))

    entitiesList = []
    for i in range(len(startIdx)-1):
        entityAtom = taggedIT[startIdx[i]: startIdx[i+1]]
        # string = ''.join([cit[0] for cit in entityAtom])
        start, end = entityAtom[0][0], entityAtom[-1][0] + 1
        tag = entityAtom[0][1].split('-')[0]
        entitiesList.append((start, end, tag))
    return entitiesList


# def extractEmbedPath2Info(embed_path, channel = None):
#     if not os.path.isfile(embed_path):
#         return None
#     path_comp = embed_path.split('/')
#     TokenNum_Dir = '/'.join(['channel'] + path_comp[1:4])
#     # print(path_comp[-1].split('.')[0].split('_')[-1].lower())
#     if channel:
#         assert channel == path_comp[-1].split('.')[0].split('_')[-1].lower()
#     else:
#         channel = path_comp[-1].split('.')[0].split('_')[-1].lower()

#     channel_abbr = CHANNEL_ABBR[channel]
#     channel_name_abbr = [i for i in path_comp[4].split('_') if channel_abbr in i][0]

#     MN_E = channel_name_abbr[len(channel_abbr): ]

#     if MN_E == '':
#          channel_name = channel
#     elif channel in CONTEXT_IND_CHANNELS:
#         channel_name = channel + MN_E
#     else:
#         if int(MN_E) == 5:
#             channel_name = channel + '-bioes'
#         elif int(MN_E) == 4:
#             channel_name = channel + '-bioe'
#         else:
#             print('Fail to Extract information for embed:', embed_path, channel, MN_E)

#     return TokenNum_Dir, channel_name





import os
import re
import numpy as np 
import pickle
import json
import pandas as pd
from datetime import datetime
from smart_open import smart_open
import itertools

from .anno import getSSET_from_CIT
from .infrastructure import strQ2B, fileReader, any2unicode

##################################################################################################CORPUS-FOLDER
# Important One 
def CorpusGroupsReader(CORPUSPath, iden = 'Dir'):
    # file is the priority
    if iden != 'Dir':
        corpusFiles = [i for i in os.listdir(CORPUSPath) if iden in i]
        if corpusFiles == []:
            results = [x for x in os.walk(CORPUSPath) if x[2]]
            results = {i[0]: i[2] for i in results}
            corpusFiles = sum([[pre + '/' + i for i in results[pre]] for pre in results], [])
            return {fd: ''      for fd in corpusFiles}, 'File'
        return {os.path.join(CORPUSPath, fd): ''      for fd in corpusFiles}, 'File'
    else:
        results = [x for x in os.walk(CORPUSPath) if x[2]]
        return {i[0]: i[2] for i in results}, 'Dir'

        
def geneTextFilePaths(corpusPath, orig_iden = '.txt', anno_iden = None):
    FolderNames = [i for i in np.sort(os.listdir(corpusPath)) if i[0] != '.']
    # print(FolderNames)
    FolderDict = {}
    
    for foldername in FolderNames:
        path = corpusPath + foldername
        # TODO: check the path by os
        OrigFileList = [i for i in os.listdir(path) if orig_iden in i ]
        if anno_iden:
            AnnoFileList = [i.replace(orig_iden, anno_iden)  for i in OrigFileList]
            AnnoFileList = [i if os.path.isfile(path + '/' + i) else '' for i in AnnoFileList]
        else:
            AnnoFileList = [''] * len(OrigFileList)
            
        FolderDict[foldername] = OrigFileList, AnnoFileList
    return FolderDict

##################################################################################################FOLDER-TEXT
anno = 'annofile4text'
anno_keywords = {
    'ANNOIden': '.NER',
    'anno_sep' : '\t', 
    'notZeroIndex' : 1, 
    'notRightOpen' : 0,
}

def annofile4text(strText, folderPath, origTextName, fileNames, ORIGIden, ANNOIden, anno_sep = '\t', notZeroIndex = 1, notRightOpen=0, **kwargs):
    SSETText = []
    name = origTextName.replace(ORIGIden, '')
    annoFileNames4Text = [f for f in fileNames if name in f and ANNOIden in f]
    
    if sum(['-sent' in f for f in annoFileNames4Text]) == 0:
        # each file stores text annotation
        strAnnoText = ''
        annoTextName = origTextName.replace(ORIGIden, ANNOIden)
        if annoTextName in fileNames:
            strAnnoText  = fileReader(os.path.join(folderPath, annoTextName))
        else:
            print('Error', os.path.join(folderPath, annoTextName))

        SSETText = [sset.split(anno_sep)[-4:] for sset in strAnnoText.split('\n') if anno_sep in sset]
        L = []
        for sset in SSETText:
            try:
                L.append([sset[0], int(sset[1]) - notZeroIndex, int(sset[2]) + notRightOpen, sset[3]])     
            except:
                pass
        SSETText = L
    return SSETText, annoTextName


anno = 'annofile4sent'
anno_keywords = {
    'ANNOIden': '.UMLSTag',
    'anno_sep' : '\t', 
    'notZeroIndex' : 0, 
    'notRightOpen' : 0,
}

def annofile4sent(strText, folderPath, origTextName, fileNames, ORIGIden, ANNOIden, anno_sep = '\t', notZeroIndex = 1, notRightOpen=0, **kwargs):
    name = origTextName.replace(ORIGIden, '')
    annoFileNames4Text = [f for f in fileNames if name in f and ANNOIden in f]
    # sentence based
    SSETText = []

    strAnnoText = ''
    sentId   = 0
    annoTextName = []
    for _ in range(len(annoFileNames4Text)):
        annoSentName = origTextName.replace(ORIGIden, '-sent' + str(sentId) + ANNOIden)
        while not annoSentName in annoFileNames4Text:
            print(sentId)
            print(annoSentName)
            sentId = sentId + 1
            annoSentName = origTextName.replace(ORIGIden, '-sent' + str(sentId) + ANNOIden)
        strAnnoText  = strAnnoText + '\n' + fileReader(os.path.join(folderPath, annoSentName))
        annoTextName.append(annoSentName)
        # print(os.path.join(folderPath, annoSentName))
        sentId = sentId + 1
        # print(strAnnoSent)
        
    SSETText = [sset.split(anno_sep)[-4:] for sset in strAnnoText.split('\n') if anno_sep in sset]
    SSETText = [[sset[0], int(sset[1]) - notZeroIndex, int(sset[2]), sset[3]] for sset in SSETText] 

    ### something different
    txtCharIdx = 0
    collapse   = 0
    for ssetIdx, sset in enumerate(SSETText):
        string, s, e, t = sset
        string = string.replace('@', ' ') # TODO
        lenString = len(string)
        while string != strText[txtCharIdx: txtCharIdx + lenString]:
            txtCharIdx = txtCharIdx + 1

        SSETText[ssetIdx] = [string, txtCharIdx, txtCharIdx + lenString, t ]
        txtCharIdx = txtCharIdx + lenString 

    return SSETText, annoTextName

def textFileReader(folderPath, fileNames, anno = False, **kwargs):
    ORIGIden = '.txt'
    # ANNOIden = anno
    origTextName = None
    annoTextName = None

    origTextNames = [f for f in fileNames if ORIGIden in f]
    for origTextName in origTextNames:
        
        # here we process each text, and want to get strText, SSET, orig, anno
        SSETText = []
        path = os.path.join(folderPath, origTextName)
        with open(path, 'r', encoding = 'utf-8') as f:
            try:
                strText = strQ2B(f.read())
            except:
                strText = ''
                print('Error in file:', path)

        if anno == 'annofile4text':
            SSETText, annoTextName = annofile4text(strText, folderPath, origTextName, fileNames, ORIGIden, **kwargs)

        elif anno == 'annofile4sent':
            SSETText, annoTextName = annofile4sent(strText, folderPath, origTextName, fileNames, ORIGIden, **kwargs)

        yield strText, SSETText, origTextName, annoTextName



anno = 'anno_embed_in_text'
anno_keywords = {}

def anno_embed_in_text(line):
    SSETText = []
    strText = ''
    ST = [(block, 'O') if idx%2==0 else (block.split(':')[-1].strip(), block.split(':')[0]) 
         for idx, block in enumerate(line.replace("}}", '{{').split('{{'))]
    txtCharIdx = 0
    for st in ST:
        string, tag = st
        strText = strText + string
        sset = [string, txtCharIdx, txtCharIdx + len(string), tag]
        txtCharIdx = sset[2]
        if tag == 'O':
            continue
        SSETText.append(sset) 
    return strText, SSETText


# 其实/o 非/o 汉/o 非/o 唐/o ，/o 又是/o 什么/o 与/o 什么/o 呢/o ？/o 
anno = 'anno_embed_along_token' 
anno_keywords = {
    'sep_between_tokens': ' ',
    'sep_between_token_label': '/', 
}

def anno_embed_along_token(line, **anno_keywords):
    line = line.replace('\n', '')
    sep_between_tokens = anno_keywords['sep_between_tokens']
    sep_between_token_label = anno_keywords['sep_between_token_label']
    tokenlabel_seq = line.split(sep_between_tokens)
    tokenlabel_split_seq = [i.split(sep_between_token_label) for i in tokenlabel_seq if sep_between_token_label in i]
    tokenlabel_split_seq = [i for i in tokenlabel_split_seq if len(i) == 2 ]
    tokenlabel_split_seq = [i for i in tokenlabel_split_seq if i[0] != '' ]
    # print(tokenlabel_split_seq)
    strText = ''
    SSETText = []
    for tokensfrag_label in tokenlabel_split_seq:
        try:
            tokensfrag, label = tokensfrag_label
        except:
            print(tokensfrag_label)
            continue
        startidx = len(strText)
        strText = strText + tokensfrag
        endidx = len(strText)
        if label.lower() != 'o':
            SSETText.append([tokensfrag, startidx, endidx, label])

    return strText, SSETText


def textLineReader(folderPath, fileNames, anno = False, **anno_keywords):
    with smart_open(folderPath) as fin:
        for line in itertools.islice(fin, None):
            line = strQ2B(any2unicode(line))
            strText = line
            SSETText = []
            if anno == 'anno_embed_in_text':
                strText, SSETText = anno_embed_in_text(line)
            elif anno == 'anno_embed_along_token':
                strText, SSETText = anno_embed_along_token(line, **anno_keywords)
            # print(strText)
            # print(SSETText)
            yield strText, SSETText, None, None

anno = 'conll_block'
anno_keywords = {
    'anno_sep': '\t',
    'connector': '',
    'suffix': False,
    'change_tags': False, # If False, B, I, I, I, ...; If True: I, I, B, ...
}

def textBlockReader(folderPath, fileNames, anno = 'conll_block', change_tags = False, 
                    connector = '', suffix = True, anno_sep = ' ', **kwargs):
    assert anno == 'conll_block'
    if suffix:
        indicator, labelidx = -1, 0
    else:
        indicator, labelidx = 0, -1
    with smart_open(folderPath) as f:
        L = []
        for line in f:
            line = strQ2B(any2unicode(line))
            if line != '\n':
                L.append(strQ2B(line).replace('\n', '').split(anno_sep)) # TODO: maybe different seps
            else:
                strText = [ct[0] for ct in L]
                strText = connector.join(strText)
                
                # print(CIT)
                # this is used to deal with conll-2003 only.
                if change_tags:
                    TotalCIT = [[ct[0], idx, ct[-1]] for idx, ct in enumerate(L)]
                    # print(TotalCIT)
                    for idx, cit in enumerate(TotalCIT):
                        currentTag = cit[-1]
                        lastTag = TotalCIT[idx - 1][-1] if idx > 1 else 'O'
                        if 'I-' not in currentTag:
                            continue
                        # lastTag = TotalCIT[idx - 1][-1] if idx > 1 else 'null'
                        if lastTag == 'O' and 'I-' in currentTag:
                            TotalCIT[idx][-1] = cit[-1].replace('I-', 'B-')
                    L = TotalCIT

                CIT = [[ct[0], idx, ct[-1]] for idx, ct in enumerate(L) if ct[-1] != 'O']

                startIdxes = [idx for idx in range(len(CIT)) if CIT[idx][-1][indicator] in ['B', 'S']] + [len(CIT)]
                # print(startIdxes)
                SSETText = []
                for i in range(len(startIdxes)-1):
                    OneCIT = CIT[startIdxes[i]: startIdxes[i+1]]
                    string = connector.join(cit[0] for cit in OneCIT)
                    start, end = OneCIT[0][1], OneCIT[-1][1] + 1
                    tag = OneCIT[0][-1].split('-')[labelidx]
                    SSETText.append([string, start, end, tag])
                # SSETText = [[sset,sset,sset,sset[3].split('-')[0]] for sset in SSETText]
                # print('In Block Reader')
                # print(strText)
                # print(SSETText)
                yield strText, SSETText, None, None
                L = []
    
# no annotation currently
def textElementReader(folderPath, fileNames, anno = False, **kwargs):
    with open(folderPath, 'rb') as handle:
        L = pickle.load(handle)
        for strText in L:
            strText = strQ2B(strText)
            yield strText, None, None, None

# for json annotation
anno = 'json_annotation'
anno_keywords = {
    'strText': 'content',
    'labels': 'annotation',
}
def textJsonReader(folderPath, fileNames, anno = 'json_annotation', strText = 'content', labels = 'annotation', **kwargs):
    assert anno == 'json_annotation'
    with open(folderPath, 'r') as file:
        for line in file.readlines():
            dic = json.loads(line)
            strText = dic[strText]
            SSETText = []
            labels = dic[labels]
            for i in labels:
                sset = [i['points'][0]['text'], i['points'][0]['start'], i['points'][0]['end']+1, ''.join(i['label'])]
                SSETText.append(sset)
            yield strText, SSETText, None, None

anno = 'csv_annotation'
anno_keywords = {
    'ANNOIden': '.xlsx',
}
def textCSVReader(folderPath, fileNames, anno = 'csv_annotation',  ANNOIden = '.csv', **kwargs):
    assert anno == 'csv_annotation'
    if anno == '.xlsx' or anno == '.csv':
        if anno == '.xlsx': 
            data = pd.read_excel(folderPath)
        if anno == '.csv':
            data = pd.read_csv(folderPath)
        strText = []
        SSETText = []
        flag = data['Sentence #'].isnull()
        for i in range(len(data['Word'])):
            if flag[i] == False and i != 0:
                strText.append(' ')
                SSETText.append([' ', -1, 'T2S'])
            word = data['Word'][i]
            label = data['Tag'][i]
            tag = [word, i, label]
            SSETText.append(tag)
            strText.append(word)
    else:       
        with open (folderPath, 'r') as file:
            lines = file.readlines()
            strText = []
            SSETText = []
            idx = 0
            for line in lines:
                if line == '\n':
                    strText.append(' ')
                    SSETText.append([' ', -1, 'T2S'])
                else:
                    line = line.split('\n')[0]
                    temp = line.split('\t')
                    strText.append(temp[0])
                    SSETText.append([temp[0], idx, temp[-1]])
                    idx += 1
    yield strText, SSETText, None, None



FolderTextsReaders = {
    'file': textFileReader,
    'line': textLineReader,
    'block':textBlockReader,
    'element': textElementReader,
    'json' : textJsonReader,
    'csv': textCSVReader
}



##################################################################################################TEXT-SENT

def reChnCutText2Sent(text, useSep = False):
    # first, eventually, the space will be used to seperate the tokens
    # in this case, if there are spaces in the original text and is not the delimiter
    # they will be removed away.

    # the formats of the text
    # 1. text is a Chinese string, in this case there are no spaces
    #    '汉字表示的数学一词大约产生于中国宋元时期。'
    #    in this case, if there is a space, regard it as an original seperator in the future.
    #    the annoying case if the sentence together with numbers and English letters.
    # 2. text is a Chinese string seperated by space
    #    '汉字 表示 的 数 学 一 词 大约 产生 于 中国 宋元 时期 。'
    # 3. tokens in the sentence contain hyper information, but still separated by space.
    #    '汉字_nz 表示_v 的_u 数_n 学_n 一_m 词_n 大约_d 产生_v 于_p 中国_ns 宋元_t 时期_n 。_w'

    ###################### Remove some weird chars #######################
    text = re.sub('\xa0', '', text)
    text = re.sub('\\n', '', text)

    # text = re.sub(r'(?<=[A-Za-z])\s+(?=[|A-Za-z])', 'ⴷ',  text)
    
    ###################### Convert the S+ spaces to '〰' #################
    # text = re.sub(' {2}', '〰', text ).strip()
    # if useSep == ' ':
    #     # if using space to sep the words
    #     text = text.replace('\t','').replace('〰', ' ')
    # elif useSep == '\t':
    #     # if using tab to sep the words, removing all spaces
    #     text = text.replace(' ','').replace('〰', '')
    # else:
    #     # if there is no sep char for Chinese, remove single space, and then convert space+ to single space
    #     text = text.replace('\t','').replace(' ', '',).replace('〰', ' ')

    text = text.replace('\t','')# .replace(' ', '',).replace('〰', ' ')
    # convert the spaces between English letters to single spaces
    # text = text.replace('ⴷ', ' ')
    
    text = re.sub('([。！;；])([^”])',  r"\1\n\2",text) 
    text = re.sub('(\.{6})([^”])',     r"\1\n\2",text) 
    text = re.sub('(\…{2})([^”])',     r"\1\n\2",text)
    
    # The \n within " " is not considered
    text = '"'.join( [ x if i % 2 == 0 else x.replace('\n', '') for i, x in enumerate(text.split('"'))] )
    
    text = re.sub( '\n+', '\n', text ).strip() # replace '\n+' to '\n'
    text = text.replace('\\n', '\n')
    text = text.split("\n")
    sents = [sent.strip() for sent in text]
    # text = [sent.replace('\\n', '') for sent in text]

    return sents # [sent for sent in text if len(sent)>=2]

def lineCutText2Sent(text):
    return text.split('\n')


def tokenText2Sent(text):
    # how to understand this?
    sents = []
    idx = 0
    for i in range(len(text)):
        if text[i] == ' ':
            sents.append(text[idx:i])
            idx =  i + 1
    sents.append(text[idx:])
    return sents

def preprocess_text(text):
    # for example: 
    # 1 removing the no-chinese character
    # 2 removing the empty parenthesis
    # 3 replacing the weird punctunations.
    return text

def segText2Sents(text, method = 'whole', **kwargs):
    # the formats of the text
    # 1. text is a Chinese string, in this case there are no spaces
    #    '汉字表示的数学一词大约产生于中国宋元时期。'
    #    in this case, if there is a space, regard it as an original seperator in the future.
    #    the annoying case if the sentence together with numbers and English letters.
    #    to deal with the spaces in this format, we simply keep them.
    # 2. text is a Chinese string seperated by space
    #    '汉字 表示 的 数 学 一 词 大约 产生 于 中国 宋元 时期 。'
    # 3. tokens in the sentence contain hyper information, but still separated by space.
    #    '汉字_nz 表示_v 的_u 数_n 学_n 一_m 词_n 大约_d 产生_v 于_p 中国_ns 宋元_t 时期_n 。_w'
    # 2. and 3. are basically the same. but we wouldn't consider it.

    # for English
    # 1. the text is naturally seperated by the spaces. we seg the text to sents by its only '\n' only.

    # TODO: here should be some checks for texts.

    if os.path.isfile(text):
        text = fileReader(text)
            
    # preprocessing the text
    text = preprocess_text(text)

    if method == 'whole':
        # this is commonly used for wikipedia data
        # for English corpus, we use this only.
        # sents = [text.replace('\\n', '').replace('\n', '').replace(' ', '').replace('\t', '').replace('\xa0', '')]
        sents = [text.replace('\\n', '').replace('\n', '').replace('\t', '')]

    elif method == 're':
        # re still need more method to consider the final results.
        sents = reChnCutText2Sent(text, **kwargs)

    elif method == 'line':
        sents = lineCutText2Sent(text)

    elif method == 'token':
        # what's the meaning?
        sents = tokenText2Sent(text)

    else:
        sents = method(text, **kwargs)

    # postprocessing
    # after this, there is no '\n' in sents
    sents = [sent.replace('\n', '').replace('\\n', '') for sent in sents]
    sents = [sent for sent in sents if len(sent) > 0]
    # print(sents)
    return sents

##################################################################################################SENT-TOKEN

def clean_token(token):
    # token contains no spaces.
    # what is token is a white splace itself?
    token = token.replace(' ', '').replace('\n', '').replace('\t', '')

    # punctuation must be seperated away from token.
    # if token contains special marks, seperate this token to two or more token.

    # last check, token should not be '', if so, pass it.
    # remove some special tokens.
    return [token]


def segSent2Tokens(sent, seg_method = 'iter', tokenLevel = 'char', Channel_Dep_Methods = {}):
    # print(Channel_Dep_Methods)
    hyper_info = {}

    # seg_method is: 1 iter; 2 ' ' or '\t'; 3 'pos' or other channel that are in Channel_Dep_Methods
    if seg_method == 'iter' and tokenLevel == 'char':
        # sent can be
        # 1. '产生于中国宋元时期。'           --->   '产 生 于 中 国 宋 元 时 期 。' 
        # 2. '产生 于 中国 宋元 时期 。'      --->   '产 生 于 中 国 宋 元 时 期 。' 
        # 3. '产 生 于 中 国 宋 元 时 期 。'  --->   '产 生 于 中 国 宋 元 时 期 。' 
        strTokens = [i for i in sent if i != ' ']
    
    elif seg_method in [' ', '\t']:
        # sent can be
        # 1. '产生 于 中国 宋元 时期 。'      --->   '产生 于 中国 宋元 时期 。' 
        # 2. '产 生 于 中 国 宋 元 时 期 。'  --->   '产 生 于 中 国 宋 元 时 期 。'
        # both word and char level is OK.
        strTokens =  [i.replace(' ', '') for i in sent.split(seg_method) if i != ' ']

        # check and clean the token
        final_tokens = []
        for idx, token in enumerate(strTokens):
            new_tokens = clean_token(token)
            final_tokens.extend(new_tokens)
        strTokens = final_tokens

    elif seg_method in Channel_Dep_Methods:
        # print(Channel_Dep_Methods[seg_method])
        ch_grain_sent, strTokens = Channel_Dep_Methods[seg_method](sent, tokenLevel = tokenLevel, tagScheme = 'BIOES')
        # check and clean the token
        final_tokens = []
        new_grain_sent = []
        for idx, token in enumerate(strTokens):
            new_tokens = clean_token(token)
            final_tokens.extend(new_tokens)
            if len(new_tokens) == 1:
                new_grain_sent.append(ch_grain_sent[idx])
            else:
                new_grain_sent.extend([ch_grain_sent[idx]] * len(new_tokens))

        strTokens = final_tokens
        ch_grain_sent = new_grain_sent
        hyper_info[seg_method] = ch_grain_sent

    else:
        raise ValueError('No good segmentaion method...')


    # remove ' ' inside strToken
    # in this case, ' ' will not be a token, we will solve this problem in the future.
    # how about, in Chinese, the ' ' between English words.
    # TODO

    # in general, we only use it for Chinese Char, and won't for Chinese Word
    # actually, we can also use it for English Word. but English char is too verbose.
    # print(strTokens)
    for ch, hyper_method in Channel_Dep_Methods.items():
        if ch == seg_method:
            continue
        # these hyper fields are all fields with larger granularity, and changing will be taken care inside the method.
        ch_grain_sent, _ = hyper_method(strTokens, tokenLevel = tokenLevel, tagScheme = 'BIOES') 
        # this assert should be inside the hyper_field_method
        # assert len(ch_grain_sent) == len(final_tokens)
        hyper_info[ch] = ch_grain_sent
        assert len(strTokens) == len(ch_grain_sent)

    return strTokens, hyper_info

#########################################################################33
def get_line_with_position(path, start_position):
    with open(path, 'r', encoding = 'utf-8') as f:
        f.seek(start_position)
        line = f.readline()
    return line[:-1]

def get_lines_with_position(path, start_position, num_lines):
    lines = ''
    i = 0
    with open(path, 'r', encoding = 'utf-8') as f:
        f.seek(start_position)
        for line in f.readlines():
            lines = lines + line 
            i = i + 1
            if i >= num_lines:
                break
    return lines[:-1]

def read_file_chunk_string(path, startbyteidx, endbyteidx):
    with open(path, 'r',  encoding = 'utf-8') as fin:
        fin.seek(startbyteidx)
        data = ''
        while True:
            data = data + fin.readline()
            # fin.tell() equal to endbyteidx is very rigid
            # we can use >= instead.
            if fin.tell() >= endbyteidx:
                break
    return data[:-1]



# for f in *_tag.txt; do mv -- "$f" "${f//_tag.txt/.NER}"; done

# ########### BOSON ###########
# CORPUSPath = 'corpus/boson/'
# corpusFileIden = '.txt'
# textType   = 'line'
# Text2SentMethod  = 're'
# Sent2TokenMethod = 'iter'
# TOKENLevel = 'char'
# anno = 'embed'
# annoKW = {}


# ########### NER ###########
# CORPUSPath = 'corpus/ner/'
# corpusFileIden = None
# textType   = 'file'
# Text2SentMethod  = 're'
# Sent2TokenMethod = 'iter'
# TOKENLevel = 'char'
# anno = '.Entity'
# annoKW = {
#     'sep': '\t',
#     'notZeroIndex': 1,
# }


# ########### MedPOS ###########
# CORPUSPath = 'corpus/medpos/'
# textType   = 'file'
# corpusFileIden = None
# Text2SentMethod  = 're'
# Sent2TokenMethod = 'iter'
# TOKENLevel = 'char'
# anno = '.UMLSTag'
# annoKW = {
#     'sep': '\t',
#     'notZeroIndex': 0,
# }

# ########### Weibo Test ###########
# CORPUSPath = 'corpus/weibotest/'
# corpusFileIden = None
# textType   = 'file'
# Text2SentMethod  = 're'
# Sent2TokenMethod = 'sep-\t'
# TOKENLevel = 'word'
# anno = False
# annoKW = {}


# ########### Wiki ###########
# CORPUSPath = 'corpus/wiki/'
# corpusFileIden = '.txt'

# textType   = 'line'

# Text2SentMethod  = 're'
# Sent2TokenMethod = 'sep- '
# TOKENLevel = 'word'

# anno = False
# annoKW = {}



# ########### ResumeNER ###########
# CORPUSPath = 'corpus/ResumeNER/'
# corpusFileIden = '.bmes'
# textType   = 'block'
# Text2SentMethod  = 're'
# Sent2TokenMethod = 'iter'
# TOKENLevel = 'char'
# anno = 'embed' # TODO
# annoKW = {}




import os
import time 
import optparse 
import pickle
from pprint import pprint
from datetime import datetime
import optparse 

import pandas as pd
import numpy as np

from .tagger import tagger
from .loaddata import loadData
from .crftools import get_sent_strfeats, get_sent_vecfeats, generate_template, crf_learn, crf_test
from .evals import get_sent_annoSET, match_anno_pred_result, calculate_F1_Score, logErrors


def crfpp_train(model, trainSents, Channel_Settings, feat_type = 'str'):
    if 'str' in feat_type.lower():
        get_sent_feats = get_sent_strfeats
    elif 'vec' in feat_type.lower():
        get_sent_feats = get_sent_vecfeats
    
    # prepare paths to save model and feats
    tmp_dir = model.replace('model', '_tmp')
    if not os.path.exists(tmp_dir): os.makedirs(tmp_dir)
    if not os.path.exists(model): os.makedirs(model)
    model_path = model + '/model'
    template_path = model + '/template'
    feats_data_path = tmp_dir + '/feats.txt'
    
    # prepare features information
    DFtrain = []
    for idx, sent in enumerate(trainSents):
        if idx % 500 == 0:
            print(datetime.now(), idx)
        df = get_sent_feats(sent, Channel_Settings)
        df.loc[len(df)] = np.NaN     ## Trick Here
        DFtrain.append(df)           ## Trick Here
    DFtrain = pd.concat(DFtrain).reset_index(drop=True)
    DFtrain.to_csv(feats_data_path, sep = '\t', encoding = 'utf=8', header = False, index = False )

    # generate template
    generate_template(input_feats_num = DFtrain.shape[1] - 1, path = template_path) 

    # train the model
    crf_learn(feats_data_path, model_path, template_path  = template_path)

    return None


def crfpp_test(model, testSents, Channel_Settings, labels):
    '''
        sents: a list of sents
        This function could be updated in the future for a better performance.
        But currently, just leave it now.
    '''
    pred_entities = []
    anno_entities = []
    log_list      = []
    # here sents are a batch of sents, not necessary to be all the sents
    # this could be chanage, but it will cause some time, so just ignore it.
    for idx, sent in enumerate(testSents):
        if idx % 200 == 0:
            print(datetime.now(), idx)
        # 200/13s
        pred_SET = tagger(model, sent, Channel_Settings = Channel_Settings)
        anno_SET = get_sent_annoSET(sent)
        error = logErrors(sent, pred_SET, anno_SET)
        log_list.append(error)

        pred_entities.append(pred_SET)
        anno_entities.append(anno_SET)


    # return anno_entities, pred_entities
    Result = match_anno_pred_result(anno_entities, pred_entities, labels = labels)
    # return Result
    R = calculate_F1_Score(Result, labels)

    LogError = pd.concat(log_list).reset_index(drop = True)
    return R, LogError


# def trainModel(model, sentTrain, sentTest, Channel_Settings, labels, cross_idx = 0):
#     '''
#         sentTrain, sentTest: are featurized already.
#     '''
#     model = model + '_' + str(cross_idx)
#     log_path = model + '/log.csv'
#     pfm_path = model + '/performance.csv'
#     para   = crfpp_train(model, sentTrain, Channel_Settings)
#     R, Err = crfpp_test (model, sentTest,  Channel_Settings, labels)
#     Err.to_csv(log_path, index = False, sep = '\t')
#     R.to_csv  (pfm_path, index = True,  sep = '\t')
#     # generate the error log path
#     return R

# def train(model, sents, Channel_Settings, labels, cross_num, cross_validation = None, seed = 10):
#     '''
#         sent is featurized
#     '''
#     total_pfm_path = model + '_performance.csv'

#     if not cross_validation:
#         sentTrain, sentTest = loadData(sents, cross_num, seed = seed, cross_validation = cross_validation, cross_idx = 0)
#         Performance = trainModel(model, sentTrain, sentTest, Channel_Settings, labels)
#         print('\nThe Final Performance is:\n')
#         print(Performance)
#         Performance.to_csv  (total_pfm_path, index = True,  sep = '\t')
#         return Performance
#     else:
#         L = []
#         for cross_idx in range(cross_num):
#             sentTrain, sentTest = loadData(sents, cross_num, seed = seed, cross_validation = cross_validation, cross_idx = cross_idx)
#             print('For  ', cross_idx, '/', cross_num, "  ====")
#             R = trainModel(model, sentTrain, sentTest, Channel_Settings, labels, cross_idx = cross_idx)
#             print(R)
#             L.append(R)
#         Performance = sum(L)/cross_num
#         print('\nThe Final Average Performance for', cross_num, 'Cross Validation is:\n')
#         print(Performance)
#         Performance.to_csv  (total_pfm_path, index = True,  sep = '\t')
#         return Performance


import numpy as np
import collections
from datetime import datetime
from .channel import getChannelGrain4Token
from .infrastructure import UNK, UNK_ID, specialTokens, specialTokensDict
import pandas as pd
##################################################################################################LTU_LGU-LT
from datetime import datetime


##################################################################################################TOKEN_LTU
def buildTokens(tokenList):
    """
        'deprecated: not suitable to handle big corpus'
        Process raw inputs into a dataset.
        words: a list of the whole corpus
    """
    #########################################################################COUNT
    total_len_token = len(tokenList)
    print('The Total Number of Tokens:', total_len_token)
    print('Counting the number unique Tokens...          \t', datetime.now())
    count = collections.Counter(tokenList).most_common()
    print('\t\tDone!')
    #########################################################################COUNT

    print('Generating Dictionary of Token Unique...\t', datetime.now())
    # grain2index
    # DTU_freq = {sp_tk: 0 for sp_tk in specialTokens}
    DTU = {}
    index2freq = []
    for token, freq in count:
        DTU[token] = len(DTU)
        index2freq.append(freq)
        # if token is not specialTokens:
        #     DTU[token] = len(DTU)
        #     DTU_freq[token] = freq
        # else:
        #     DTU_freq[token] = DTU_freq[token] + 1

    print('\t\tThe length of original vocabulary is:', len(DTU), '\t', datetime.now())
    print('Generating the ORIGTokenIndex...       \t', datetime.now())
    data = np.zeros(len(tokenList), dtype = np.uint32)
    # data = []
    for idx, token in enumerate(tokenList):
        data[idx] = DTU[token]
        if idx % 5000000 == 0:
            print('\t\tThe idx of token is:', idx, '\t', datetime.now())
    print('\t\tDone!')
    LTU = list(DTU.keys())
    # if MaxTokenUnique:
    #     print('Only Keep First', MaxTokenUnique, 'Tokens.')
    #     print('The coverage rate is:', np.bincount(data)[UNK_ID]/total_len_token)
    # data = np.array(data)
    return data, LTU, DTU, index2freq
##################################################################################################TOKEN_LTU




def get_num_freq(idx2freq, min_token_freq = 1):
    max_vocab_token_num = len(idx2freq[idx2freq >= min_token_freq])
    return max_vocab_token_num

def get_GU_or_LKP(TokenVocab, tkidx2freq, 
                  channel= 'char', Min_Ngram = 1, Max_Ngram = 1, end_grain = False, min_grain_freq = 1):

    # ListGrainUnique = []
    LTU, DTU = TokenVocab
    # max_vocab_token_num = get_num_freq(tkidx2freq, min_token_freq = min_token_freq)
    # LTU = LTU[:max_vocab_token_num]
    
    # the containers to store our results
    oldLGU = []
    oldDGU = {}
    oldidx2freq = []
    LKP = []
    
    print('For channel: |', channel, '| build GrainUnique and LookUp')
    for idx, token in enumerate(LTU):
        token_freq  = tkidx2freq[DTU[token]]
        ChN = getChannelGrain4Token(token, channel, Min_Ngram = Min_Ngram, Max_Ngram = Max_Ngram, end_grain = end_grain)
        grain2number = dict(collections.Counter(ChN).most_common())
        for gr in grain2number:
            if gr in oldDGU:
                oldidx2freq[oldDGU[gr]] = oldidx2freq[oldDGU[gr]] + grain2number[gr] * token_freq
                # oldidx2freq[oldDGU[gr]] = oldidx2freq[oldDGU[gr]] + grain2number[gr] 
            else:
                oldDGU[gr] = len(oldDGU)
                oldLGU.append(gr)
                oldidx2freq.append(grain2number[gr] * token_freq)
                # oldidx2freq.append(grain2number[gr])

        LKP.append([oldDGU[gr] for gr in ChN])
        if idx % 100000 == 0:
            print('\t\tFor Channel:', channel, '\t', idx, datetime.now())

    # remove some high and low frequency grains.
    # how to deal with the high freqency grains?
    # notice that the grain freq is based on vocab instead of corpus.
    del oldDGU 
    assert len(LKP) == len(LTU)
    
    # sort the LGU, DGU and renew LKP
    oldidx2freq = np.array(oldidx2freq)
    max_grain_num = len(oldidx2freq[oldidx2freq >= min_grain_freq])
    
    grainidx2freq = np.sort(oldidx2freq)[::-1]
    newidx2oldidx = np.argsort(oldidx2freq)[::-1]
    del oldidx2freq

    oldidx2newidx = np.zeros(len(newidx2oldidx), dtype= int) 
    for new_idx, old_idx in enumerate(newidx2oldidx):
        oldidx2newidx[old_idx] = new_idx
    
    for tkidx, grainlist in enumerate(LKP):
        new_grainlist = []
        for oldidx in grainlist:
            newidx = oldidx2newidx[oldidx]
            # throw away the low frequency grains
            if grainidx2freq[newidx] < min_grain_freq:
                continue
            new_grainlist.append(newidx)
        LKP[tkidx] = new_grainlist 
    del oldidx2newidx

    LGU = []
    for new_idx in range(max_grain_num):
        # to filter some grains
        LGU.append(oldLGU[newidx2oldidx[new_idx]])
    del oldLGU
    del newidx2oldidx

    DGU = {}
    for new_idx, token in enumerate(LGU):
        DGU[token] = new_idx
        
    grainidx2freq = grainidx2freq[:max_grain_num]
    
    return (LGU, DGU), LKP, grainidx2freq


# UPDATE: 2018/03/14, Add concise argument.
# UPDATE: 2018/05/04, For Linxu.
import os
import pickle
import numpy as np
import pandas as pd
from io import StringIO
from .nlptext_lite.utils.channel import getChannelName

console_encoding = 'gb2312'
file_encoding = 'utf-8'


def load_anno(BasicObject, anno_field):
    # channel_anno = 'annoE'
    Channel_Settings = BasicObject.CHANNEL_SETTINGS
    tagScheme = Channel_Settings[anno_field]['tagScheme']
    GU = BasicObject.getGrainVocab(anno_field, tagScheme = tagScheme)
    tags = GU[0]
    tag_size = len(tags)
    labels = list(set([i.split('-')[0] for i in tags if '-' in i]))
    labels.sort()
    return Channel_Settings, tagScheme, labels, tags, tag_size


def get_model_name(BasicObject, Channel_Settings):
    return BasicObject.Data_Dir.replace('data/', 'model/') + '/' + "_".join([getChannelName(ch, style = 'abbr') 
                                            for ch in Channel_Settings])

def dict2list(paramdict):
    resultlist = []
    for k, v in paramdict.items():
        resultlist.append(k)
        if v: resultlist.append(v)
    return resultlist

def shell_invoke(args, sinput = None, soutput = None):
    import subprocess
    if sinput and soutput:
        p = subprocess.Popen(args, stdin = sinput, stdout= soutput)
    elif sinput:
        p = subprocess.Popen(args, stdin=sinput, stdout=subprocess.PIPE)
    elif soutput:
        p = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=soutput)
    else:
        p = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    result = p.communicate()
    for robj in result:
        if robj:
            print(robj.decode(console_encoding))
    return None


learn_params  = {'-f': '2',
                 '-c': '5.0',
                 '-t': None}

def crf_learn(train_data_path, model_path,
              template_path  = 'template/template01',
              crf_learn_path = 'crfpp/source/crf_learn',
              params = learn_params):
    '''
    Use train data from `train_data_path` to learn a model and save the model to `model_path`
    You may need to specify template_path or crf_learn_path
    '''
    part_args = []
    if params:
        part_args += dict2list(params)
    part_args += [template_path, train_data_path, model_path]
    try:
        shell_invoke([crf_learn_path] + part_args)
    except:
        shell_invoke(['crf_learn'] + part_args)

def crf_test(test_data_path, model_path, result_path, 
             concise = True, crf_test_path = 'crfpp/source/crf_test'):
    '''
    Use test data from `test_data_path` and the model from `model_path` to save the result in `result_path`
    You may need to specify the concise or crf_test_path
    '''
    if not concise: 
        part_args = ['-v', '2', '-m', model_path, test_data_path]
        with open(result_path, 'w') as fh_write:
            try:
                shell_invoke([crf_test_path] + part_args, soutput = fh_write)
            except:
                shell_invoke(['crf_test'] + part_args, soutput = fh_write)
    else:
        part_args = ['-m', model_path, test_data_path]
        
        with open(result_path, 'w') as fh_write:
            try:
                shell_invoke([crf_test_path] + part_args, soutput = fh_write)
            except:
                shell_invoke(['crf_test'] + part_args, soutput = fh_write)


# Dev in 2019/06/14


############################################################################## prepare feats

Total_Settings = {
    'token': {'Max_Ngram': 1},
    'char': {'Max_Ngram': 1, 'end_grain': False},
    'basic': {'Max_Ngram': 1, 'end_grain': False},
    'radical': {'Max_Ngram': 1, 'end_grain': False},
    'subcomp': {'Max_Ngram': 1, 'end_grain': False},
    'stroke': {'Max_Ngram': 1, 'end_grain': False},
    'pos': {'Max_Ngram': 1, 'end_grain': False,   'tagScheme': 'BIOE'},
    'annoE': {'Max_Ngram': 1, 'end_grain': False, 'tagScheme': 'BIOE'}
}

def get_sent_strfeats(sent, Channel_Settings, train = True):
    '''
        sent is a nlptex.sentence object
        return a pandas dataframe
    '''
    try:
        df = sent.feats
        columns = df.columns
        new_columns = [i for i in columns if i.split('_')[0] in Channel_Settings]
        if not train:
            new_columns = [i for i in new_columns if 'anno' not in i]
        Feats = df.loc[new_columns]
        return Feats
    except:
        features = {}
        # stroke 12 and subcomp 6 are fixed internally.
        for ch, cs in Channel_Settings.items():
            if 'anno' in ch and not train:
                continue
            feature = sent.get_grain_str(ch)
            # this will cost a lot of time
            if ch == 'stroke':
                max_leng = 12
                feature2 = []
                for token_feat in feature:
                    if len(token_feat) <= max_leng:
                        feature2.append(token_feat + ['</'] * (max_leng - len(token_feat))) 
                    else:
                        feature2.append(token_feat[:max_leng])
                feature = feature2
            elif ch == 'subcomp':
                max_leng = 6
                feature2 = []
                for token_feat in feature:
                    if len(token_feat) <= max_leng:
                        feature2.append(token_feat + ['</'] * (max_leng - len(token_feat)))
                    else:
                        feature2.append(token_feat[:max_leng])
                feature = feature2 
            features[ch] = feature
            # print(feature)
        L = []
        for ch, feat in features.items():
            df = pd.DataFrame(feat)
            df.columns = [ch + '_' + str(i) for i in range(df.shape[1])]
            L.append(df)
        
        Feats = pd.concat(L, axis = 1)
        return Feats

def get_sent_vecfeats(sent, fieldembed, train = True):
    features = {}
    Channel_Settings = fieldembed.Field_Settings
    for ch, cs in Channel_Settings.items():
        if 'anno' in ch and not train:
            continue
        derivative_wv = fieldembed.weigth[ch].derivative_wv
        LTU = derivative_wv.LGU # LGU in derivative wv is LTU
        token_idxes = sent.get_grain_idx('token', LTU = LTU)
        feature = derivative_wv.vectors[token_idxes]
        features[ch] = feature
        # print(feature)
    return 

def featurize_nlptext_sentences(BasicObject, feat_type = 'str', fieldembed = None):
    if 'str' in feat_type.lower():
        sentence_path = BasicObject.TokenNum_Dir + '/' + 'Pyramid/_Feat_SENT_Str.crfpp'
        get_sent_feats = get_sent_strfeats
    elif 'vec' in feat_type.lower():
        sentence_path = BasicObject.TokenNum_Dir + '/' + 'Pyramid/_Feat_SENT_Vec.crfpp'
        get_sent_feats = get_sent_vecfeats
    else:
        print('Error! Currently there is no such a feature type')
        
    if os.path.isfile(sentence_path):

        with open(sentence_path, 'rb') as handle:
            sents = pickle.load(handle)
        print('Load Featurized Sentences from:')
        print('\t', sentence_path)
        return sents
    else:
        from nlptext.corpus import Corpus
        corpus = Corpus() # this costs time
        sents = []
        for sent in corpus.Sentences:
            sent.feats = get_sent_feats(sent, Total_Settings)
            sents.append(sent)
        with open(sentence_path, 'wb') as handle:
            pickle.dump(sents, handle)
        print('Save Featurized Sentences to:')
        print('\t', sentence_path)
        return sents

##############################################################################


############################################################################## generate template for derivative features

# individaul
individual_para = {
    1: 0,
    # 2: 2,
    # 3: 1,
}

def generate_template(input_feats_num = 5, individual_para = individual_para, path = None):
    '''
    this still needs more consideration
    '''
    if not path:
        path = '_template'
        
    L = ['# Unigram\n\n']
    fld_idx = 0
    for feat_idx in range(input_feats_num):
        for gram_num, window_size in individual_para.items():
            if gram_num == 1:
                for token_i in range(-window_size, window_size + 1):
                    L.append('U{}:%x[{},{}]\n'.format(str(fld_idx), str(token_i), str(feat_idx)))
                    fld_idx = fld_idx + 1
            if gram_num == 2 and feat_idx <= 5:
                for token_i in range(-window_size, window_size + 1):
                    L.append('U{}:%x[{},{}]/%x[{},{}]\n'.format(str(fld_idx), 
                                                                str(token_i), str(feat_idx), 
                                                                str(token_i + 1), str(feat_idx)))
                    fld_idx = fld_idx + 1
                    
            if gram_num == 3 and feat_idx <= 5:
                for token_i in range(-window_size, window_size + 1):
                    L.append('U{}:%x[{},{}]/%x[{},{}]/%x[{},{}]\n'.format(str(fld_idx), 
                                                                          str(token_i - 1), str(feat_idx), 
                                                                          str(token_i), str(feat_idx), 
                                                                          str(token_i + 1), str(feat_idx)))
                    fld_idx = fld_idx + 1
    
    
    L.append('\n\n# Bigram\nB')
    
    with open(path, 'w') as f:
        f.write(''.join(L))
        
    return ''.join(L)



from os import path
from io import open
from setuptools import setup, find_packages

HERE = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(HERE, 'README.md'), encoding='utf-8') as f:
    LONG_DESCRIPTION = f.read()

with open(path.join(HERE, 'requirements.txt'), encoding='utf-8') as f:
    INSTALL_REQUIRES = f.read().split('\n')

# # Get VERSION from the VERSION file
# with open(path.join(HERE, 'VERSION'), encoding='utf-8') as f:
#     PACKAGE_VERSION = f.read().rstrip()

PACKAGE_VERSION = '0.0.0.1'

setup(
    name='medpos',

    version=PACKAGE_VERSION,

    description='medical part-of-speech tagging',

    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    
    package_data={  # Optional
        'medpos': ['crfpp/nlptext_lite/sources/*',
                    'model/MedPos/char/T_b-n1t1-f1_m-n1t1-f1_r-n1t1-f1_P-bio_E-bio/model',
                    'model/MedPos/char/T_b-n1t1-f1_m-n1t1-pf1_r-n1t1-f1_P-bio_E-bio/para.p',
                    'model/MedPos/char/T_b-n1t1-f1_m-n1t1-f1_r-n1t1-f1_P-bio_E-bio/template'],
    },

    url='',

    author='junjieluo',

    author_email='floydjjluo@gmail.com',

    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish
        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    # This field adds keywords for your project which will appear on the
    # project page. What does your project relate to?
    #
    # Note that this is a string of words separated by whitespace, not a list.
    keywords='',  # Optional

    packages=find_packages(exclude=[
        'contrib',
        'docs',
        ]),  # Required

    python_requires='>=3.6, <4',

    install_requires=[] + INSTALL_REQUIRES,  # Optional

    extras_require={  # Optional
        'dev': ['check-manifest'],
        'test': [
            'coverage',
            'pytest',
            'pylint',
        ],
    },

    
)



import os
import re
import pickle
import string
from jieba import posseg
import nltk

import pyphen
from .infrastructure import specialTokens

################## FOR THE CONTEXT-INDEPENDENT CHANNELS ################

##### basic
def basicGrainChar(char, end_grain = False):
    '''char level only!'''
    punStr = string.punctuation + '＂＃＄％＆＇（）＊＋，－／：；＜＝＞＠［＼］＾＿｀｛｜｝～｟｠｢｣､\u3000、〃〈〉《》「」『』【】〔〕〖〗〘〙〚〛〜〝〞〟〰〾〿–—‘’‛“”„‟…‧﹏﹑﹔·！？｡。'
    engReg = r'[A-Za-z]{1}'
    if '%' in char or '%' in char:
        info =  'PERC'
    elif re.match(r'[0-9]{1}', char):
        info = "NUM"
    elif char in punStr:
        info = "PUNC"
    elif char >= '\u4e00' and char <= '\u9fff':
        info = "CHN"
    elif re.match(engReg, char):
        info = 'ENG'
    #elif word in string.whitespace:
        #return 'SPA'
    elif char == '@':
        info = 'SPA'
    else:
        info = 'OTHER'
    info = [info]
    if end_grain:
        info = info + ['b0']
    return info

def basicGrainToken(token, end_grain = False):
    info = sum([basicGrainChar(char, end_grain) for char in token], [])
    return info

#### medical

def medicalGrainChar(word, end_grain = False):
    '''char level only!'''
    units = 'kBq kbq mg Mg UG Ug ug MG ml ML Ml GM iu IU u U g G l L cm CM mm s S T % % mol mml mmol MMOL HP hp mmHg umol ng'.split(' ')
    chn_units = '毫升 毫克 单位 升 克 第 粒 颗粒 支 件 散 丸 瓶 袋 板 盒 合 包 贴 张 泡 国际单位 万 特充 个 分 次'.split(' ')
    med_units = 'qd bid tid qid qh q2h q4h q6h qn qod biw hs am pm St DC prn sos ac pc gtt IM IV po iH'.split(' ')
    all_units = units + chn_units + med_units

    site_units = '上 下 左 右 间 片 部 内 外 前 侧 后'.split(' ')
    sym_units = '大 小 增 减 多 少 升 降 高 低 宽 厚 粗 两 双 延 长 短 疼 痛 终 炎 咳'.split(' ')
    part_units = '脑 心 肝 脾 肺 肾 胸 脏 口 腹 胆 眼 耳 鼻 颈 手 足 脚 指 壁 膜 管 窦 室 管 髋 头 骨 膝 肘 肢 腰 背 脊 腿 茎 囊 精 唇 咽'.split(' ')
    break_units = '呈 示 见 伴 的 因'.split(' ')
    more_units = '较 稍 约 频 偶 偏'.split(' ')
    non_units = '无 不 非 未 否'.split(' ')
    tr_units = '服 予 行'.split(' ')

    if word in units:
        info =  'UNIT'
    elif word in chn_units:
        info =  'CHN_UNIT'
    elif word in med_units:
        info =  'MED_UNIT'
    elif word in site_units:
        info =  'SITE_UNIT'
    elif word in sym_units:
        info =  'SYM_UNIT'
    elif word in part_units:
        info =  'PART_UNIT'
    elif word in break_units:
        info =  'BREAK_UNIT'
    elif word in more_units:
        info =  'more_UNIT'
    elif word in non_units:
        info =  'NON_UNIT'
    elif word in tr_units:
        info =  'TR_UNIT'
    else:
        info =  'OTHER'
        
    # return [info]
    info = [info]
    if end_grain:
        info = info + ['m0']
    return info

def medicalGrainToken(token, end_grain = False):
    info = sum([medicalGrainChar(char, end_grain) for char in token], [])
    return info
    



current_dir_old = os.path.abspath(os.path.dirname(__file__))
current_dir, _ = os.path.split(current_dir_old)

# print(current_dir_old)
# print(current_dir)
# data_dir = os.path.join(current_dir, 'data')



#### subcomp

with open( os.path.join(current_dir, 'sources/CharSubComp.p'), 'rb') as handle:
    CharSubCompInfos = pickle.load(handle)

#### stroke

with open( os.path.join(current_dir, 'sources/CharStroke.p'), 'rb') as handle:
    CharStrokeInfos = pickle.load(handle)
    
#### radical

with open( os.path.join(current_dir, 'sources/CharRadical.p'), 'rb') as handle:
    CharRadicalInfos = pickle.load(handle)

#### pinyin

with open( os.path.join(current_dir, 'sources/CharPinyin.p'), 'rb') as handle:
    CharPinyinInfos = pickle.load(handle)


## phonemes
with open( os.path.join(current_dir, 'sources/WordPhoneme.p'), 'rb') as handle:
    WordPhenomeInfo = pickle.load(handle)


def subcompGrainChar(char, end_grain = False):
    '''char level only!'''
    if char in CharSubCompInfos:
        info = CharSubCompInfos[char]
        if info:
            info = ['c' + i for i in info ]
        else:
            info = ['c' + char] 
    else:
        info = [char]
        
    if end_grain:
        info = info + ['c0']
    return info

def subcompGrainToken(token, end_grain = False):
    info = sum([subcompGrainChar(char, end_grain) for char in token], [])
    return info


def strokeGrainChar(char, end_grain = False):
    '''char level only!'''
    if char in CharStrokeInfos:
        info = CharStrokeInfos[char]
        if info:
            info = ['s' + i for i in info ]
        else:
            info = ['s' + char] 
    else:
        info = [char]
        
    if end_grain:
        info = info + ['s0']
    return info

def strokeGrainToken(token, end_grain = False):
    info = sum([strokeGrainChar(char, end_grain) for char in token], [])
    return info

def radicalGrainChar(char, end_grain = False):
    ''''char level only!'''
    if char in CharRadicalInfos:
        info = CharRadicalInfos[char]
        if info:
            info = [info] # here Radical Data is in str not list
        else:
            info = [char]
    else:
        info = [char]
        
    if end_grain:
        info = info + ['r0']
    return info

def radicalGrainToken(token, end_grain = False):
    info = sum([radicalGrainChar(char, end_grain) for char in token], [])
    return info

#### char
def charGrainChar(char, end_grain = False):
    '''char level only!'''
    info = [char]
    if end_grain:
        info = info + ['ch0']
    return info

def charGrainToken(token, end_grain = False):
    info = sum([charGrainChar(char, end_grain) for char in token], [])
    return info


def pinyinGrainChar(char, end_grain = False):
    '''char level only!'''
    if char in CharPinyinInfos:
        info = CharPinyinInfos[char]
        if info:
            info = [i for i in info ]
        else:
            info = ['po'] 
    else:
        info = [char]
        
    if end_grain:
        info = info + ['y0']
    return info

def pinyinGrainToken(token, end_grain = False):
    info = sum([pinyinGrainChar(char, end_grain) for char in token], [])
    return info


## syllable
def syllableGrainToken(token, end_grain = False):
    # pyphen.LANGUAGES
    try:
        dic = pyphen.Pyphen(lang='en')
        info = dic.inserted(token).split('-')
    except:
        info = ['']
    return info


def phonemeGrainToken(token, end_grain = False): 
    try:
        phonemes = WordPhenomeInfo[token.lower()]
    except:
        phonemes = ['']
    return phonemes

################## FOR THE CONTEXT-DEPENDENT CHANNELS ################
posTag = ['a', 'ad', 'ag', 'an', 'b', 'c', 'd', 'df', 'dg', 'e', 'eng', 
          'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'mg', 'mq', 'n', 'ng', 'nr', 'nrfg', 
          'nrt', 'ns', 'nt', 'nz', 'o', 'p', 'q', 'r', 'rg', 'rr','rz', 's', 't', 
          'tg', 'u', 'ud', 'ug', 'uj', 'ul', 
          'uv', 'uz', 'v', 'vd', 'vg', 'vi', 'vn', 'vq', 'x', 'y', 'yg', 'z', 'zg']

def POSGrainSent(sent, tokenLevel = 'word', tagScheme = 'BIOES'):
    '''
        # tokenLevel = 'word'
            1: sent = '北京 是 中国 的 首都' # be cautious about this sentence, final segmentation may be different from the input
            2: sent = '北京是中国的首都'     # this is ok
        Here only for Atom is Char Based
        This method should be enriched
        sent: List of Token(String), with or without Start or End
        sent: [str]
    '''
    if type(sent) == str:
        sent = sent.split(' ') 
    segs = list(posseg.cut(''.join(sent)))
    tokens = []
    GrainSent = []

    if tokenLevel == 'char':
        for i in range(len(segs)):
            pair  = segs[i]
            leng  = len(pair.word)
            label = pair.flag
            labels= [label + '-I' ]*leng
            labels[0] = label + '-B'
            if 'E' in tagScheme and leng >= 2:
                labels[-1] = label + '-E'
            if 'S' in tagScheme and leng == 1:
                labels[0] = label + '-S'
            GrainSent.extend(labels)
            tokens.extend([i for i in pair.word])
        return GrainSent, tokens

    elif tokenLevel == 'word':
        for i in range(len(segs)):
            pair  = segs[i]
            label = pair.flag
            GrainSent.append(label+ '-B')
            tokens.append(pair.word )
            # print(pair.word)
        return GrainSent, tokens



MedPosTag = ['notsure', '临床属性', '事件','人群', '代词', '体征与症状',
 '医疗仪器', '医疗行为', '单位', '因果', '地名机构名', '定性', '或许','数字', '数学符号',
 '无', '时间单位', '时间概念', '有', '标点', '检查项目', '治疗项目', '物体',
 '生物', '疾病', '空间概念', '药物', '虚词','身体功能', '身体物质',  '身体部位', '连词']




def MedPOSGrainSent(sent, tokenLevel = 'char', tagScheme = 'BIOES'):
    

    result_entities = medpos_tagger(sent)
    sentence = [i for i in sent]

    entities = []
    for i in result_entities:
        s, e, label = i
        entities.append((''.join(sentence[s:e]), s, e, label))
    # entities
    
  
    assert tokenLevel == 'char'
    tokens = []
    GrainSent = []

    try:
        for sset in entities:
            string, s, e, label = sset
            leng  = len(string)
            labels= [label + '-I' ]*leng
            labels[0] = label + '-B'
            if 'E' in tagScheme and leng >= 2:
                labels[-1] = label + '-E'
            if 'S' in tagScheme and leng == 1:
                labels[0] = label + '-S'
            GrainSent.extend(labels)
            tokens.extend([i for i in string])
    except:
        print(entities)


    # print(list(zip(tokens, GrainSent)))
    
    return GrainSent, tokens




posEnTag = ['LS', 'TO', 'VBN', "''", 'WP', 'UH', 'VBG', 'JJ', 'VBZ', '--', 
            'VBP', 'NN', 'DT', 'PRP', ':', 'WP$', 'NNPS', 'PRP$', 'WDT', 
            '(', ')', '.', ',', '``', '$', 'RB', 'RBR', 'RBS', 'VBD', 'IN', 
            'FW', 'RP', 'JJR', 'JJS', 'PDT', 'MD', 'VB', 'WRB', 'NNP', 'EX',
            'NNS', 'SYM', 'CC', 'CD', 'POS']

def POSENGrainSent(sent, tokenLevel = 'word', tagScheme = 'BIOES'):
    '''
        Here only for Atom is Char Based
        This method should be enriched
        sent: List of Token(String), with or without Start or End
        sent: [str]
    '''
    assert tokenLevel == 'word'
    if type(sent) == list:
        tokens = sent
    else:
        tokens = nltk.word_tokenize(sent)
    # segs = list(posseg.cut(''.join(sent)))
    GrainSent = [i[-1] +'-B' for i in nltk.pos_tag(tokens)]
    return GrainSent, tokens


##################################################################

################################################################################################
CONTEXT_IND_CHANNELS    = ['basic', 'medical', 'radical', 'token', 'char', 'subcomp', 'stroke', 'pinyin', 'syllable', 'phoneme', ]
CONTEXT_DEP_CHANNELS    = ['pos', 'pos_en', 'medpos']
ANNO_CHANNELS           = ['annoR', 'annoE']

CONTEXT_IND_CHANNELS_AB = ['b', 'm', 'r',  'T', 'C', 'c', 's', 'y', 'sl', 'ph']
CONTEXT_DEP_CHANNELS_AB = ['P', 'Pe', 'MP']
ANNO_CHANNELS_AB        = ['R', 'E']

CHANNEL_ABBR = dict(zip(CONTEXT_IND_CHANNELS + CONTEXT_DEP_CHANNELS+ANNO_CHANNELS , 
                        CONTEXT_IND_CHANNELS_AB+CONTEXT_DEP_CHANNELS_AB + ANNO_CHANNELS_AB ))


Channel_Ind_Methods ={
    'char': charGrainToken, 
    'basic': basicGrainToken,
    'medical':medicalGrainToken,
    'radical':radicalGrainToken,
    'subcomp':subcompGrainToken,
    'stroke':strokeGrainToken,
    'pinyin':pinyinGrainToken,
    'syllable': syllableGrainToken,
    'phoneme': phonemeGrainToken, 
}

Channel_Dep_Methods = {'pos': POSGrainSent, 
                       'pos_en': POSENGrainSent, 
                       'medpos': MedPOSGrainSent}

Channel_Dep_TagSets = {'pos': posTag, 
                       'pos_en': posEnTag, 
                       'medpos': MedPosTag}


################################################################################################

def getGrainNgrams(subword_infos, n):
    if n == 1:
        return [i for i in subword_infos]
    if n > len(subword_infos):
        # How to deal this when the length is not so long
        # Condition: where n is larger than the infos
        return [] 
    l = [subword_infos[i:n+i] for i in range(len(subword_infos) - n + 1)]
    l = ['-'.join(i) for i in l]
    return l


def grainToken(token, grainTokenFunction, Ngram = None, Min_Ngram = 1, Max_Ngram = 1, end_grain = True):
    infos =  grainTokenFunction(token, end_grain = end_grain) 
    if Ngram: Min_Ngram, Max_Ngram = Ngram, Ngram
    return sum([getGrainNgrams(infos, idx) for idx in range(Min_Ngram, Max_Ngram + 1)], [])


def getChannelGrain4Token(token, channel, Ngram = None, Min_Ngram = 1, Max_Ngram = 1,  end_grain = False):
    if channel == 'token':
        return [token]
    elif channel in Channel_Ind_Methods:
        return grainToken(token, Channel_Ind_Methods[channel], Ngram = Ngram, Min_Ngram = Min_Ngram,Max_Ngram = Max_Ngram, end_grain = end_grain)
    else:
        print('The Channel "', channel, '" is not available currently!')


def getChannelGrain4Sent(sent, channel, Ngram = None, Min_Ngram = 1, Max_Ngram = 1, tokenLevel = 'char', tagScheme =  'BIO', end_grain = False):
    if channel == 'token':
        return [[tk] for tk in sent.split(' ')]
    elif channel in Channel_Ind_Methods:
        return [getChannelGrain4Token(token, channel, Ngram, Min_Ngram, Max_Ngram, end_grain) for token in sent.split(' ')]
        # return grainSent_ctxInd(sent, channel, Ngram = Ngram, Max_Ngram = Max_Ngram,  end_grain = end_grain)
    elif channel in Channel_Dep_Methods:
        return [[gr] for gr in Channel_Dep_Methods[channel](sent, tokenLevel = tokenLevel, tagScheme = tagScheme)[0]]
        # return grainSent_ctxDep(sent, Channel_Dep_Methods[channel], tokenLevel =tokenLevel, tagScheme = tagScheme, useStartEnd = useStartEnd)
    else:
        print('The Channel "', channel, '" is not available currently!')


###############################################################################################################

###############################################################################################################
def getChannelName(channel, Min_Ngram = 1, Max_Ngram = 1,  end_grain = False, tagScheme = 'BIO', min_grain_freq = 1,
                   style = 'normal', channel_name = None, channel_name_abbr = None, **kwargs):
    # hyper: channel + '-bio'
    # sub  : channel + '-n1t5' + 'e' or '' + 'f5'
    if style == 'normal':
        if channel == 'token':
            return channel 
        elif channel in CONTEXT_IND_CHANNELS:
            MN = '-n' + str(Min_Ngram) + 't' + str(Max_Ngram)
            e  = 'e' if end_grain else ''
            f  = '-f' + str(min_grain_freq)
            return channel + MN + e + f
        else:
            tS = '-' + tagScheme.lower() 
            return channel + tS

    elif style == 'abbr':
        channel_abbr = CHANNEL_ABBR[channel] 
        if channel == 'token':
            return channel_abbr
        elif channel in CONTEXT_IND_CHANNELS:
            MN = '-n' + str(Min_Ngram) + 't' + str(Max_Ngram)
            e  = 'e' if end_grain else ''
            f  = '-f' + str(min_grain_freq)
            return channel_abbr + MN + e + f
        else:
            tS = '-' + tagScheme.lower() 
            return channel_abbr + tS

    elif channel_name and style == 'extract':
        assert channel in channel_name
        if channel == 'token':
            return channel, Min_Ngram, Max_Ngram, end_grain, min_grain_freq, tagScheme

        elif channel in CONTEXT_IND_CHANNELS:
            channel_name, freq = channel_name.split('-f')
            min_grain_freq = int(freq)
            MN_e = channel_name.split('-n')[1]
            # print(MN_e)
            if 'e' in MN_e:
                end_grain = True
                MN = MN_e[:-1]
            else:
                end_grain = False
                MN = MN_e
            Min_Ngram, Max_Ngram = [int(i) for i in MN.split('t')]
            return channel, Min_Ngram, Max_Ngram, end_grain, min_grain_freq, tagScheme
        
        else:
            tagScheme = channel_name.split('-')[-1].upper()
            return channel, Min_Ngram, Max_Ngram, end_grain, min_grain_freq, tagScheme
        
    elif channel_name_abbr and style == 'extract':
        channel_abbr = CHANNEL_ABBR[channel]
        assert channel_abbr in channel_name_abbr
        if channel == 'token':
            return channel_abbr, Min_Ngram, Max_Ngram, end_grain, min_grain_freq, tagScheme

        elif channel in CONTEXT_IND_CHANNELS:
            channel_name, freq = channel_name_abbr.split('-f')
            min_grain_freq = int(freq)
            MN_e = channel_name.split('-n') [1]
            if 'e' in MN_e:
                end_grain = True
                MN = MN_e[:-1]
            else:
                end_grain = False
                MN = MN_e
            Min_Ngram, Max_Ngram = [int(i) for i in MN.split('t')]
            return channel_abbr, Min_Ngram, Max_Ngram, end_grain, min_grain_freq, tagScheme
        
        else:
            tagScheme = channel_name.split('-')[-1].upper()
            return channel_abbr, Min_Ngram, Max_Ngram, end_grain, min_grain_freq, tagScheme
    else:
        print('Error in getChannelName')


def get_Channel_Settings(CHANNEL_SETTINGS_TEMPLATE):
    d = CHANNEL_SETTINGS_TEMPLATE.copy()
    try:
        CHANNEL_SETTINGS = {channel: d[channel] for channel in d 
                            if d[channel].pop('use') == True}
    except:
        CHANNEL_SETTINGS = {k:v for k, v in d.items()}

    nameList = []

    for channel in CHANNEL_SETTINGS:
        # CHANNEL_SETTINGS[]
        channel_setting = CHANNEL_SETTINGS[channel]
        #channel_setting.pop('use')
        Max_Ngram    = channel_setting.get('Max_Ngram', 1)
        end_grain    = channel_setting.get('end_grain', False)
        tagSet       = channel_setting.get('tagSet',    None)
        tagScheme    = channel_setting.get('tagScheme', 'BIO')

        channel_name_abbr = getChannelName(channel, Max_Ngram, end_grain, tagScheme, style = 'abbr')
        nameList.append(channel_name_abbr)
    
    folderName = '_'.join(nameList)
    
    return CHANNEL_SETTINGS, folderName
    


import pandas as pd
import numpy as np
from tabulate import tabulate

###################################################  from IT to SET 

def get_target_field_info(nlptext, Target_Field, fldembed, useStartEnd = False):
    field, para = Target_Field
    # nlptext is important here.
    # channel_anno = 'annoE'
    target_para = {}
    
    if field == 'token':
        # use start and end only works for token
        special_num = 3 if useStartEnd else 1 
        # from fldembed
        Weights = fldembed.weights
        assert 'token' in Weights
        wv = Weights['token']
        GU = wv.GU
        idx2grain = ['</pad>'] + GU[0]
        grain_num = len(idx2grain)
        grain2idx = dict(zip(idx2grain, list(range(grain_num)) ))
        GU = idx2grain, grain2idx
        target_para['GU']  = GU # </pad> in GU, no </unk> in GU
        labels = [] 
        tag_size = len(idx2grain) + special_num # </pad> in GU, and </unk>, </start>, </end> not in GU

    else: 
        # generally, we won't adding start and end into CRF model.
        # even the features are with start and end, it will should be removed before feed into crf
        # it can also be proved that the start and end vectors are useless in sequential labeling.
        GU = nlptext.getGrainVocab(field, **para)
        idx2tag = ['</pad>'] + GU[0]
        tag2idx = dict(zip(range(len(idx2tag)), idx2tag))
        GU = idx2tag, tag2idx
        # </pad> is in GU, no </unk> or other special tags in GU
        tag_size = len(idx2tag)       
        target_para['GU'] = GU
        
        TRANS = nlptext.getTrans(field, **para)
        TRANS = {k: v+1 for k, v in TRANS.items()}
        labels = list(set([i.split('-')[0] for i in idx2tag if '-' in i]))
        labels.sort()
        target_para['TRANS'] = TRANS
        
    # target_para['labels'] = labels
    # target_para['tag_size'] = tag_size
    
    TARGET_FIELD = [field, target_para, labels, tag_size]
    return TARGET_FIELD



def read_target_seq(result_path):
    result = pd.read_csv(result_path, sep = '\t', header = None, skip_blank_lines=False)
    # get the last column of the result
    return result.iloc[:,-1].dropna().values


def extractSET(tag_seq, exist_SE = False):
    '''
        SET: start, end, tag
        tag_seq: the hyper field sequence for this sentence

    '''
    if exist_SE:
        tag_seq = tag_seq[1:-1]

    #################################### TODO
    IT = list(zip(range(len(tag_seq)), tag_seq))
    taggedIT = [list(it) for it in IT if it[1]!= 'O']
    first_tag = taggedIT[0][1].split('-')[0]
    
    taggedIT[0][1] = first_tag + '-B'
    
    startIdx = [idx for idx in range(len(taggedIT)) if taggedIT[idx][1][-2:] == '-B' or taggedIT[idx][1][-2:] == '-S']
    startIdx.append(len(taggedIT))

    entitiesList = []
    for i in range(len(startIdx)-1):
        entityAtom = taggedIT[startIdx[i]: startIdx[i+1]]
        # string = ''.join([cit[0] for cit in entityAtom])
        start, end = entityAtom[0][0], entityAtom[-1][0] + 1
        tag = entityAtom[0][1].split('-')[0]
        entitiesList.append((start, end, tag))
    return entitiesList


def get_sent_annoSET(sent, channel = 'annoE', tagScheme = 'BIO'):
    anno_seq = [i[0] for i in sent.get_grain_str(channel, tagScheme=tagScheme)]
    anno_SET = extractSET(anno_seq)
    return anno_SET
################################################### 



###################################################  compare pred_SET and anno_SET

def match_anno_pred_result(anno_entities, pred_entities, labels = []):
    '''
        anno_entities, pred_entities of a batch of sentences
    '''
    if type(anno_entities[0]) != list:
        anno_entities = [anno_entities]
        pred_entities = [pred_entities]
        
    name_list = ['E@Anno', 'E@Pred',  'E@Match']
    for eL in labels:
        name_list.extend([eL + suff for suff in ['@Anno', '@Pred', '@Match']])
    
    statistic_result = []
    
    for idx in range(len(pred_entities)):
        pred = set(pred_entities[idx])
        anno = set(anno_entities[idx])
        d = {'E@Pred'  : len(pred),
             'E@Anno'  : len(anno),
             'E@Match' : len(pred.intersection(anno))}
        
        for eL in labels:
            elL = [e for e in pred if eL == e[-1]]
            elA = [e for e in anno if eL == e[-1]]
            elM = set(elA).intersection(set(elL)) ## Union vs Join
            d[eL+'@Pred']  = len(elL)
            d[eL+'@Anno']  = len(elA)
            d[eL+'@Match'] = len(elM)
        
        statistic_result.append(d)
    Result = pd.DataFrame(statistic_result)[name_list]
    return Result



def calculate_F1_Score(Result, labels):
    if len(Result.shape) == 1:
        Result = Result.to_dict()
    else:
        Result = Result.sum().to_dict()
    List = []
    entitiesLabel = labels + ['E']
    # entitiesLabel = ['Sy','Bo', 'Ch', 'Tr', 'Si'] + ['R'] + ['E']
    for eL in entitiesLabel:
        d = dict()
        d['id'] = eL
        for k in Result:
            if eL == k.split('@')[0]:
                d[k.replace(eL + '@','')] = Result[k]
        List.append(d)
    
    # print(List)
    R = pd.DataFrame(List)
    R.set_index('id', inplace = True)
    R.index.name = None
    # print(R)
    R['R'] = R['Match']/R['Anno']
    R['P'] = R['Match']/R['Pred']
    R['F1'] = 2*R['R']*R['P']/(R['R'] + R['P'])
    return R[['Anno', 'Pred', 'Match', 'R', 'P', 'F1']]


################################################### log the errors between pred_SET and anno_SET
def matchPaired(L, A, sent):
    start1, end1, e1 = L
    start2, end2, e2 = A
    d = {}
    sentence = sent.sentence.split(' ')
    if set(range(start1, end1+1)).intersection(range(start2, end2+1)):
        idx = set(range(start1, end1+1)).union(range(start2, end2+1))
        # print()
        d['text_part'] = ''.join(sentence[min(idx): max(idx)])
        d['start'] = min(idx)
        d['end' ]  = max(idx) 
        d['pred'] = ''.join(sentence[start1: end1])
        d['pred_en'] = e1
        d['anno'] = ''.join(sentence[start2: end2])
        d['anno_en'] = e2
        d['sent_idx']= sent.Idx # this is important
        return d
    
def matchUnpaired(unpaired, sent, kind):
    d = {}
    sentence = sent.sentence.split(' ')
    d['start'], d['end' ], e = unpaired
    d['text_part'] = ''.join(sentence[d['start']: d['end' ]])
    d['sent_idx']= sent.Idx
    if kind == 'P':
        d['pred'], d['pred_en'] = d['text_part'], e
        d['anno'], d['anno_en'] = None, None
    else:
        d['pred'], d['pred_en'] = None, None
        d['anno'], d['anno_en'] = d['text_part'], e
    return d


def logErrors(sent, anno_enetities, pred_entities):
    log = []
    inter = list(set(pred_entities).intersection(set(anno_enetities)))
    only_pred = [ i for i in pred_entities if i not in inter]   
    only_anno = [ i for i in anno_enetities if i not in inter]
    
    pairedP = []
    pairedA = []
    for L in only_pred:
        for A in only_anno:
            d = matchPaired(L, A, sent)
            if d:
                log.append(d)
                pairedP.append(L)
                pairedA.append(A)
                
    for L in [i for i in only_pred if i not in pairedP]:
        log.append(matchUnpaired(L, sent, 'P'))
        
    for A in [i for i in only_anno if i not in pairedA]:
        log.append(matchUnpaired(A, sent, 'A'))
           
    if len(log) == 0:
        return pd.DataFrame()
    cols = ['sent_idx', 'text_part', 'start', 'end', 'anno', 'anno_en', 'pred', 'pred_en']
    return pd.DataFrame(log)[cols].sort_values('start')

